# This file is a part of Groebner.jl. License is GNU GPL v2.

# Custom assertions for Groebner. Provides the @invariant macro.

"""
    @invariant expr

Check that `expr` evaluates to `true` at runtime. If not, throw an
`AssertionError`.

It should be fine to use the `@invariant` macro in performance-critical code, as
long as the code is compiled with `invariants_enabled() = false`.

## Examples

```jldoctest
@invariant 2 > 1
@invariant 1 > 2  # throws!
```
"""
macro invariant(arg)
    # NOTE: does not propagate the error source path
    expr = quote
        if $(@__MODULE__).invariants_enabled()
            @assert $arg
        else
            nothing
        end
    end
    esc(expr)
end

macro unreachable()
    :(@invariant false)
end

# Invokes undefined behavior in LLVM. 
# Only use when able to prove that this is never getting called.
unsafe_unreachable() = Base.llvmcall("unreachable", Cvoid, Tuple{})
