using AbstractAlgebra
using Primes

@testset "groebner modular-hard problems" begin
    R, (x1, x2, x3, x4) = polynomial_ring(QQ, ["x1", "x2", "x3", "x4"], ordering=:degrevlex)

    ###########################################################################

    N = prod(map(BigInt, nextprimes(2^30 + 3, 5)))
    system = [
        x1 + x2 + x3 + x4,
        x1 * x2 + x1 * x3 + x1 * x4 + x2 * x3 + x2 * x4 + x3 * x4,
        x1 * x2 * x3 + x1 * x2 * x4 + x1 * x3 * x4 + x2 * x3 * x4,
        x1 * x2 * x3 * x4 + N
    ]

    # this should take about 10 primes
    gb = Groebner.groebner(system)
    @test gb == [
        x1 + x2 + x3 + x4,
        x2^2 + x2 * x3 + x3^2 + x2 * x4 + x3 * x4 + x4^2,
        x3^3 + x3^2 * x4 + x3 * x4^2 + x4^3,
        x4^4 - N
    ]

    ###########################################################################

    N = prod(map(BigInt, nextprimes(2^31 - 1, 5)))
    system = [
        x1 + x2 + x3 + x4,
        x1 * x2 + x1 * x3 + x1 * x4 + x2 * x3 + x2 * x4 + x3 * x4,
        x1 * x2 * x3 + x1 * x2 * x4 + x1 * x3 * x4 + x2 * x3 * x4,
        x1 * x2 * x3 * x4 + N
    ]

    # this should take about 10 primes
    gb = Groebner.groebner(system)
    @test gb == [
        x1 + x2 + x3 + x4,
        x2^2 + x2 * x3 + x3^2 + x2 * x4 + x3 * x4 + x4^2,
        x3^3 + x3^2 * x4 + x3 * x4^2 + x4^3,
        x4^4 - N
    ]

    ###########################################################################

    N = prod(map(BigInt, nextprimes(2^31 - 1, 100)))
    system = [
        x1 + x2 + x3 + x4,
        x1 * x2 + x1 * x3 + x1 * x4 + x2 * x3 + x2 * x4 + x3 * x4,
        x1 * x2 * x3 + x1 * x2 * x4 + x1 * x3 * x4 + x2 * x3 * x4,
        x1 * x2 * x3 * x4 + N
    ]
    # around 200 primes are required
    gb = Groebner.groebner(system)
    @test gb == [
        x1 + x2 + x3 + x4,
        x2^2 + x2 * x3 + x3^2 + x2 * x4 + x3 * x4 + x4^2,
        x3^3 + x3^2 * x4 + x3 * x4^2 + x4^3,
        x4^4 - N
    ]

    ###########################################################################
    N = prod(
        map(
            BigInt,
            primes(
                2^30 +
                2^29 +
                2^28 +
                2^27 +
                2^26 +
                2^25 +
                2^24 +
                2^23 +
                2^22 +
                2^21 +
                2^20 +
                2^19 +
                2^18 +
                2^17 +
                2^16 +
                2^15 +
                2^14,
                2^31 - 1
            )
        )
    )

    # this should take A LOT of primes
    system = [
        x1 + x2 + x3 + x4,
        x1 * x2 + x1 * x3 + x1 * x4 + x2 * x3 + x2 * x4 + x3 * x4,
        x1 * x2 * x3 + x1 * x2 * x4 + x1 * x3 * x4 + x2 * x3 * x4,
        x1 * x2 * x3 * x4 + N
    ]
    # impossible to compute for now
    # UPD: now computes in 7 seconds
    # UPD 2: now computes in 1 second !!
    # UPD 3: now computes in < 0.3 second !!
    gb = Groebner.groebner(system)
    @test gb == [
        x1 + x2 + x3 + x4,
        x2^2 + x2 * x3 + x3^2 + x2 * x4 + x3 * x4 + x4^2,
        x3^3 + x3^2 * x4 + x3 * x4^2 + x4^3,
        x4^4 - N
    ]

    system = [x1 - (2^31 - 1) * x2 - (2^30 + 3) * x3]
    @test Groebner.groebner(system) == system

    ###########################################################################
    for start_of_range in [2^10, 2^20, 2^30, 2^40]
        for size_of_range in [10, 100, 1000]
            N = prod(Primes.nextprimes(BigInt(start_of_range), size_of_range))
            system = [x1 + N // (N + 1), x3 - (N + 1) // (N - 1), x2 + 42]
            gb = Groebner.groebner(system)
            @test gb == [x3 - (N + 1) // (N - 1), x2 + 42, x1 + N // (N + 1)]
        end
    end
end

@testset "groebner, hard-coded answer" begin
    noon = Groebner.noonn(4, k=QQ, ordering=:degrevlex)
    (x1, x2, x3, x4) = gens(parent(first(noon)))

    gb = Groebner.groebner(noon)

    sing_ans = [
        10 * x1^2 * x4 + 10 * x2^2 * x4 + 10 * x3^2 * x4 - 11 * x4 + 10,
        10 * x1^2 * x3 + 10 * x2^2 * x3 + 10 * x3 * x4^2 - 11 * x3 + 10,
        10 * x1 * x2^2 + 10 * x1 * x3^2 + 10 * x1 * x4^2 - 11 * x1 + 10,
        10 * x1^2 * x2 + 10 * x2 * x3^2 + 10 * x2 * x4^2 - 11 * x2 + 10,
        x3^3 * x4 - x3 * x4^3 + x3 - x4,
        x2^3 * x4 - x2 * x4^3 + x2 - x4,
        20 * x2^2 * x3^2 + 20 * x2^2 * x4^2 + 20 * x3^2 * x4^2 + 11 * x1^2 - 11 * x2^2 -
        11 * x3^2 - 11 * x4^2 - 10 * x1 +
        10 * x2 +
        10 * x3 +
        10 * x4,
        x2^3 * x3 - x2 * x3^3 + x2 - x3,
        10 * x1 * x2 * x3^2 * x4 + 20 * x1 * x2 * x4^3 - 11 * x1 * x2 * x4 - 10 * x1 * x2 +
        10 * x1 * x4 +
        10 * x2 * x4,
        200 * x2 * x3^4 + 200 * x2 * x3^2 * x4^2 + 200 * x2 * x4^4 - 110 * x2^3 -
        220 * x2 * x3^2 - 220 * x2 * x4^2 - 100 * x1 * x2 + 100 * x2^2 - 100 * x2 * x3 +
        200 * x3^2 - 100 * x2 * x4 +
        200 * x4^2 +
        121 * x2 - 110,
        200 * x1 * x3^4 + 200 * x1 * x3^2 * x4^2 + 200 * x1 * x4^4 - 110 * x1^3 -
        220 * x1 * x3^2 - 220 * x1 * x4^2 + 100 * x1^2 - 100 * x1 * x2 - 100 * x1 * x3 +
        200 * x3^2 - 100 * x1 * x4 +
        200 * x4^2 +
        121 * x1 - 110,
        20 * x1 * x2 * x3^3 + 10 * x1 * x2 * x3 * x4^2 - 11 * x1 * x2 * x3 - 10 * x1 * x2 +
        10 * x1 * x3 +
        10 * x2 * x3,
        600 * x3^2 * x4^4 + 110 * x1^4 + 110 * x2^4 - 220 * x3^4 - 660 * x3^2 * x4^2 -
        220 * x4^4 - 100 * x1^3 - 100 * x2^3 - 300 * x1 * x3^2 - 300 * x2 * x3^2 +
        200 * x3^3 - 300 * x3^2 * x4 - 300 * x1 * x4^2 - 300 * x2 * x4^2 +
        300 * x3 * x4^2 +
        200 * x4^3 - 121 * x1^2 - 121 * x2^2 +
        242 * x3^2 +
        242 * x4^2 +
        220 * x1 +
        220 * x2 - 110 * x3 - 110 * x4 - 400,
        600 * x2^2 * x4^4 + 110 * x1^4 - 220 * x2^4 + 110 * x3^4 - 660 * x2^2 * x4^2 -
        220 * x4^4 - 100 * x1^3 + 200 * x2^3 - 300 * x2^2 * x3 + 300 * x1 * x3^2 -
        100 * x3^3 - 300 * x2^2 * x4 + 300 * x2 * x4^2 - 300 * x3 * x4^2 + 200 * x4^3 -
        121 * x1^2 + 242 * x2^2 - 121 * x3^2 + 242 * x4^2 - 110 * x1 - 110 * x2 +
        220 * x3 - 110 * x4 - 100,
        400 * x2 * x3^2 * x4^3 + 200 * x2 * x4^5 - 220 * x2 * x3^2 * x4 - 330 * x2 * x4^3 - 200 * x2 * x3^2 - 100 * x1 * x2 * x4 +
        100 * x2^2 * x4 +
        100 * x2 * x3 * x4 +
        200 * x3^2 * x4 - 100 * x2 * x4^2 +
        200 * x4^3 +
        121 * x2 * x4 +
        110 * x2 - 220 * x4,
        400 * x1 * x3^2 * x4^3 + 200 * x1 * x4^5 - 220 * x1 * x3^2 * x4 - 330 * x1 * x4^3 - 200 * x1 * x3^2 - 100 * x1 * x2 * x4 - 100 * x2^2 * x4 +
        100 * x1 * x3 * x4 +
        100 * x3^2 * x4 - 100 * x1 * x4^2 +
        200 * x4^3 +
        121 * x1 * x4 +
        110 * x1 - 110 * x4 - 100,
        400 * x2^2 * x3 * x4^3 + 200 * x3 * x4^5 - 220 * x2^2 * x3 * x4 - 330 * x3 * x4^3 - 200 * x2^2 * x3 + 200 * x2^2 * x4 - 100 * x1 * x3 * x4 +
        100 * x2 * x3 * x4 +
        100 * x3^2 * x4 - 100 * x3 * x4^2 +
        200 * x4^3 +
        121 * x3 * x4 +
        110 * x3 - 220 * x4,
        30 * x1 * x2 * x3 * x4^3 - 11 * x1 * x2 * x3 * x4 - 20 * x1 * x2 * x3 +
        10 * x1 * x2 * x4 +
        10 * x1 * x3 * x4 +
        10 * x2 * x3 * x4,
        2200 * x3^6 - 2200 * x4^6 - 2000 * x3^5 - 4000 * x1 * x3^2 * x4^2 -
        4000 * x2 * x3^2 * x4^2 + 4000 * x3^2 * x4^3 - 8000 * x1 * x4^4 -
        8000 * x2 * x4^4 - 4000 * x3 * x4^4 + 2000 * x4^5 - 3630 * x3^4 +
        3630 * x4^4 +
        2200 * x1^3 +
        2200 * x2^3 +
        1100 * x1 * x3^2 +
        1100 * x2 * x3^2 +
        2200 * x3^3 - 3300 * x3^2 * x4 +
        7700 * x1 * x4^2 +
        7700 * x2 * x4^2 +
        3300 * x3 * x4^2 - 2200 * x4^3 - 2000 * x1^2 + 4000 * x1 * x2 - 2000 * x2^2 +
        1000 * x1 * x3 +
        1000 * x2 * x3 - 3669 * x3^2 +
        3000 * x1 * x4 +
        3000 * x2 * x4 +
        4000 * x3 * x4 - 16331 * x4^2 - 2420 * x1 - 2420 * x2 + 4400,
        2200 * x2^6 - 2200 * x4^6 - 2000 * x2^5 - 4000 * x2^2 * x3 * x4^2 +
        4000 * x1 * x3^2 * x4^2 +
        4000 * x2^2 * x4^3 - 4000 * x1 * x4^4 - 4000 * x2 * x4^4 - 8000 * x3 * x4^4 +
        2000 * x4^5 - 3630 * x2^4 +
        3630 * x4^4 +
        2200 * x1^3 +
        2200 * x2^3 +
        1100 * x2^2 * x3 - 1100 * x1 * x3^2 + 2200 * x3^3 - 3300 * x2^2 * x4 +
        2200 * x1 * x4^2 +
        3300 * x2 * x4^2 +
        7700 * x3 * x4^2 - 2200 * x4^3 - 2000 * x1^2 + 1000 * x1 * x2 - 3669 * x2^2 +
        4000 * x1 * x3 +
        1000 * x2 * x3 - 2000 * x3^2 +
        3000 * x1 * x4 +
        4000 * x2 * x4 +
        3000 * x3 * x4 - 12331 * x4^2 - 1210 * x1 - 2420 * x3 + 3300,
        2200 * x1^6 - 2200 * x4^6 - 2000 * x1^5 +
        4000 * x2^2 * x3 * x4^2 +
        4000 * x2 * x3^2 * x4^2 - 4000 * x2^2 * x4^3 - 4000 * x3^2 * x4^3 -
        4000 * x1 * x4^4 - 4000 * x2 * x4^4 - 4000 * x3 * x4^4 + 2000 * x4^5 -
        3630 * x1^4 +
        3630 * x4^4 +
        2200 * x1^3 +
        2200 * x2^3 - 1100 * x2^2 * x3 - 1100 * x2 * x3^2 +
        2200 * x3^3 +
        3300 * x2^2 * x4 +
        3300 * x3^2 * x4 +
        3300 * x1 * x4^2 +
        2200 * x2 * x4^2 +
        2200 * x3 * x4^2 +
        2200 * x4^3 - 3669 * x1^2 + 1000 * x1 * x2 - 2000 * x2^2 +
        1000 * x1 * x3 +
        4000 * x2 * x3 - 2000 * x3^2 +
        4000 * x1 * x4 +
        3000 * x2 * x4 +
        3000 * x3 * x4 - 12331 * x4^2 - 1210 * x2 - 1210 * x3 - 3630 * x4 + 5500,
        66000 * x4^7 + 180000 * x1 * x4^5 + 180000 * x2 * x4^5 + 180000 * x3 * x4^5 -
        60000 * x4^6 - 133100 * x4^5 +
        22000 * x1^4 +
        22000 * x2^4 +
        22000 * x3^4 - 242000 * x1 * x4^3 - 242000 * x2 * x4^3 - 242000 * x3 * x4^3 +
        44000 * x4^4 - 20000 * x1^3 - 20000 * x2^3 - 20000 * x3^3 -
        60000 * x1 * x2 * x4 - 60000 * x1 * x3 * x4 - 60000 * x2 * x3 * x4 -
        120000 * x1 * x4^2 - 120000 * x2 * x4^2 - 120000 * x3 * x4^2 + 509860 * x4^3 -
        36300 * x1^2 - 36300 * x2^2 - 36300 * x3^2 +
        72600 * x1 * x4 +
        72600 * x2 * x4 +
        72600 * x3 * x4 +
        24200 * x4^2 +
        88000 * x1 +
        88000 * x2 +
        88000 * x3 - 311641 * x4 - 96690,
        6000 * x3 * x4^6 - 2200 * x3^5 - 9900 * x3 * x4^4 - 4000 * x1 * x3^3 -
        4000 * x2 * x3^3 + 2000 * x3^4 - 3000 * x1 * x3 * x4^2 - 3000 * x2 * x3 * x4^2 +
        3000 * x3^2 * x4^2 - 7000 * x3 * x4^3 +
        6000 * x4^4 +
        3630 * x3^3 +
        3630 * x3 * x4^2 - 1100 * x1^2 - 1100 * x2^2 +
        3300 * x1 * x3 +
        3300 * x2 * x3 - 2200 * x3^2 + 6600 * x3 * x4 - 7700 * x4^2 +
        1000 * x1 +
        1000 * x2 - 4331 * x3 - 3000 * x4 + 1210,
        6000 * x2 * x4^6 - 2200 * x2^5 - 9900 * x2 * x4^4 +
        2000 * x2^4 +
        4000 * x1 * x2 * x3^2 - 4000 * x2 * x3^3 +
        1000 * x1 * x2 * x4^2 +
        3000 * x2^2 * x4^2 - 3000 * x2 * x3 * x4^2 - 7000 * x2 * x4^3 +
        6000 * x4^4 +
        3630 * x2^3 +
        3630 * x2 * x4^2 - 1100 * x1^2 - 1100 * x1 * x2 - 2200 * x2^2 + 3300 * x2 * x3 -
        1100 * x3^2 + 6600 * x2 * x4 - 7700 * x4^2 +
        1000 * x1 +
        3669 * x2 - 3000 * x3 - 3000 * x4 + 1210,
        6000 * x1 * x4^6 - 2200 * x1^5 - 9900 * x1 * x4^4 +
        2000 * x1^4 +
        4000 * x1 * x2 * x3^2 - 4000 * x1 * x3^3 + 1000 * x1 * x2 * x4^2 -
        3000 * x2^2 * x4^2 - 3000 * x1 * x3 * x4^2 - 3000 * x3^2 * x4^2 -
        7000 * x1 * x4^3 +
        6000 * x4^4 +
        3630 * x1^3 +
        3630 * x1 * x4^2 - 2200 * x1^2 - 1100 * x1 * x2 - 1100 * x2^2 + 3300 * x1 * x3 -
        1100 * x3^2 + 6600 * x1 * x4 - 4400 * x4^2 +
        3669 * x1 +
        1000 * x2 - 3000 * x3 - 6000 * x4 + 1210,
        600 * x2 * x3 * x4^5 - 550 * x2 * x3 * x4^3 - 200 * x2 * x3^3 -
        100 * x1 * x2 * x3 * x4 +
        100 * x2^2 * x3 * x4 +
        100 * x2 * x3^2 * x4 - 500 * x2 * x3 * x4^2 +
        400 * x2 * x4^3 +
        400 * x3 * x4^3 +
        121 * x2 * x3 * x4 +
        330 * x2 * x3 - 220 * x2 * x4 - 220 * x3 * x4 - 200 * x3 + 200 * x4,
        600 * x1 * x3 * x4^5 - 550 * x1 * x3 * x4^3 - 200 * x1 * x3^3 -
        100 * x1 * x2 * x3 * x4 - 100 * x2^2 * x3 * x4 + 100 * x1 * x3^2 * x4 -
        500 * x1 * x3 * x4^2 +
        400 * x1 * x4^3 +
        300 * x3 * x4^3 +
        121 * x1 * x3 * x4 +
        330 * x1 * x3 - 220 * x1 * x4 - 110 * x3 * x4 - 200 * x3 + 100 * x4,
        600 * x1 * x2 * x4^5 - 550 * x1 * x2 * x4^3 + 200 * x1 * x2 * x3^2 -
        100 * x1 * x2 * x3 * x4 - 100 * x1 * x3^2 * x4 - 100 * x2 * x3^2 * x4 -
        300 * x1 * x2 * x4^2 +
        300 * x1 * x4^3 +
        300 * x2 * x4^3 +
        121 * x1 * x2 * x4 +
        110 * x1 * x2 - 110 * x1 * x4 - 110 * x2 * x4
    ]

    @test gb == map(f -> divexact(f, leading_coefficient(f)), sing_ans)

    noon = Groebner.noonn(5, ordering=:degrevlex)
    (x1, x2, x3, x4, x5) = gens(parent(first(noon)))

    gb = Groebner.groebner(noon)
    sing_ans = [
        10 * x1^2 * x5 + 10 * x2^2 * x5 + 10 * x3^2 * x5 + 10 * x4^2 * x5 - 11 * x5 + 10,
        10 * x1^2 * x4 + 10 * x2^2 * x4 + 10 * x3^2 * x4 + 10 * x4 * x5^2 - 11 * x4 + 10,
        10 * x1^2 * x3 + 10 * x2^2 * x3 + 10 * x3 * x4^2 + 10 * x3 * x5^2 - 11 * x3 + 10,
        10 * x1 * x2^2 + 10 * x1 * x3^2 + 10 * x1 * x4^2 + 10 * x1 * x5^2 - 11 * x1 + 10,
        10 * x1^2 * x2 + 10 * x2 * x3^2 + 10 * x2 * x4^2 + 10 * x2 * x5^2 - 11 * x2 + 10,
        x4^3 * x5 - x4 * x5^3 + x4 - x5,
        x3^3 * x5 - x3 * x5^3 + x3 - x5,
        x2^3 * x5 - x2 * x5^3 + x2 - x5,
        x3^3 * x4 - x3 * x4^3 + x3 - x4,
        x2^3 * x4 - x2 * x4^3 + x2 - x4,
        20 * x2^2 * x3^2 +
        20 * x2^2 * x4^2 +
        20 * x3^2 * x4^2 +
        20 * x2^2 * x5^2 +
        20 * x3^2 * x5^2 +
        20 * x4^2 * x5^2 +
        11 * x1^2 - 11 * x2^2 - 11 * x3^2 - 11 * x4^2 - 11 * x5^2 - 10 * x1 +
        10 * x2 +
        10 * x3 +
        10 * x4 +
        10 * x5,
        x2^3 * x3 - x2 * x3^3 + x2 - x3,
        10 * x1 * x2 * x3^2 * x5 + 10 * x1 * x2 * x4^2 * x5 + 20 * x1 * x2 * x5^3 -
        11 * x1 * x2 * x5 - 10 * x1 * x2 +
        10 * x1 * x5 +
        10 * x2 * x5,
        10 * x1 * x2 * x3^2 * x4 + 20 * x1 * x2 * x4^3 + 10 * x1 * x2 * x4 * x5^2 -
        11 * x1 * x2 * x4 - 10 * x1 * x2 +
        10 * x1 * x4 +
        10 * x2 * x4,
        200 * x2 * x3^4 +
        200 * x2 * x3^2 * x4^2 +
        200 * x2 * x4^4 +
        200 * x2 * x3^2 * x5^2 +
        200 * x2 * x4^2 * x5^2 +
        200 * x2 * x5^4 - 110 * x2^3 - 220 * x2 * x3^2 - 220 * x2 * x4^2 -
        220 * x2 * x5^2 - 100 * x1 * x2 + 100 * x2^2 - 100 * x2 * x3 + 200 * x3^2 -
        100 * x2 * x4 + 200 * x4^2 - 100 * x2 * x5 +
        200 * x5^2 +
        121 * x2 - 110,
        200 * x1 * x3^4 +
        200 * x1 * x3^2 * x4^2 +
        200 * x1 * x4^4 +
        200 * x1 * x3^2 * x5^2 +
        200 * x1 * x4^2 * x5^2 +
        200 * x1 * x5^4 - 110 * x1^3 - 220 * x1 * x3^2 - 220 * x1 * x4^2 -
        220 * x1 * x5^2 + 100 * x1^2 - 100 * x1 * x2 - 100 * x1 * x3 + 200 * x3^2 -
        100 * x1 * x4 + 200 * x4^2 - 100 * x1 * x5 +
        200 * x5^2 +
        121 * x1 - 110,
        20 * x1 * x2 * x3^3 + 10 * x1 * x2 * x3 * x4^2 + 10 * x1 * x2 * x3 * x5^2 -
        11 * x1 * x2 * x3 - 10 * x1 * x2 +
        10 * x1 * x3 +
        10 * x2 * x3,
        200 * x2 * x3^2 * x4^2 * x5 +
        400 * x2 * x3^2 * x5^3 +
        400 * x2 * x4^2 * x5^3 +
        200 * x2 * x5^5 - 220 * x2 * x3^2 * x5 - 220 * x2 * x4^2 * x5 -
        330 * x2 * x5^3 - 200 * x2 * x3^2 - 200 * x2 * x4^2 - 100 * x1 * x2 * x5 +
        100 * x2^2 * x5 +
        100 * x2 * x3 * x5 +
        200 * x3^2 * x5 +
        100 * x2 * x4 * x5 +
        200 * x4^2 * x5 - 100 * x2 * x5^2 +
        200 * x5^3 +
        121 * x2 * x5 +
        110 * x2 - 220 * x5,
        200 * x1 * x3^2 * x4^2 * x5 +
        400 * x1 * x3^2 * x5^3 +
        400 * x1 * x4^2 * x5^3 +
        200 * x1 * x5^5 - 220 * x1 * x3^2 * x5 - 220 * x1 * x4^2 * x5 -
        330 * x1 * x5^3 - 200 * x1 * x3^2 - 200 * x1 * x4^2 - 100 * x1 * x2 * x5 -
        100 * x2^2 * x5 +
        100 * x1 * x3 * x5 +
        100 * x3^2 * x5 +
        100 * x1 * x4 * x5 +
        100 * x4^2 * x5 - 100 * x1 * x5^2 +
        200 * x5^3 +
        121 * x1 * x5 +
        110 * x1 - 110 * x5 - 100,
        200 * x2^2 * x3 * x4^2 * x5 +
        400 * x2^2 * x3 * x5^3 +
        400 * x3 * x4^2 * x5^3 +
        200 * x3 * x5^5 - 220 * x2^2 * x3 * x5 - 220 * x3 * x4^2 * x5 -
        330 * x3 * x5^3 - 200 * x2^2 * x3 - 200 * x3 * x4^2 + 200 * x2^2 * x5 -
        100 * x1 * x3 * x5 +
        100 * x2 * x3 * x5 +
        100 * x3^2 * x5 +
        100 * x3 * x4 * x5 +
        200 * x4^2 * x5 - 100 * x3 * x5^2 +
        200 * x5^3 +
        121 * x3 * x5 +
        110 * x3 - 220 * x5,
        10 * x1 * x2 * x3 * x4^2 * x5 + 30 * x1 * x2 * x3 * x5^3 - 11 * x1 * x2 * x3 * x5 - 20 * x1 * x2 * x3 +
        10 * x1 * x2 * x5 +
        10 * x1 * x3 * x5 +
        10 * x2 * x3 * x5,
        600 * x3^2 * x4^4 +
        600 * x3^2 * x4^2 * x5^2 +
        600 * x3^2 * x5^4 +
        600 * x4^2 * x5^4 +
        110 * x1^4 +
        110 * x2^4 - 220 * x3^4 - 660 * x3^2 * x4^2 - 220 * x4^4 - 660 * x3^2 * x5^2 -
        660 * x4^2 * x5^2 - 220 * x5^4 - 100 * x1^3 - 100 * x2^3 - 300 * x1 * x3^2 -
        300 * x2 * x3^2 + 200 * x3^3 - 300 * x3^2 * x4 - 300 * x1 * x4^2 -
        300 * x2 * x4^2 +
        300 * x3 * x4^2 +
        200 * x4^3 - 300 * x3^2 * x5 - 300 * x4^2 * x5 - 300 * x1 * x5^2 -
        300 * x2 * x5^2 +
        300 * x3 * x5^2 +
        300 * x4 * x5^2 +
        200 * x5^3 - 121 * x1^2 - 121 * x2^2 +
        242 * x3^2 +
        242 * x4^2 +
        242 * x5^2 +
        220 * x1 +
        220 * x2 - 110 * x3 - 110 * x4 - 110 * x5 - 500,
        600 * x2^2 * x4^4 +
        600 * x2^2 * x4^2 * x5^2 +
        600 * x2^2 * x5^4 +
        600 * x4^2 * x5^4 +
        110 * x1^4 - 220 * x2^4 + 110 * x3^4 - 660 * x2^2 * x4^2 - 220 * x4^4 -
        660 * x2^2 * x5^2 - 660 * x4^2 * x5^2 - 220 * x5^4 - 100 * x1^3 + 200 * x2^3 -
        300 * x2^2 * x3 + 300 * x1 * x3^2 - 100 * x3^3 - 300 * x2^2 * x4 +
        300 * x2 * x4^2 - 300 * x3 * x4^2 + 200 * x4^3 - 300 * x2^2 * x5 -
        300 * x4^2 * x5 + 300 * x2 * x5^2 - 300 * x3 * x5^2 +
        300 * x4 * x5^2 +
        200 * x5^3 - 121 * x1^2 + 242 * x2^2 - 121 * x3^2 +
        242 * x4^2 +
        242 * x5^2 - 110 * x1 - 110 * x2 + 220 * x3 - 110 * x4 - 110 * x5 - 200,
        400 * x2 * x3^2 * x4^3 +
        200 * x2 * x4^5 +
        200 * x2 * x3^2 * x4 * x5^2 +
        400 * x2 * x4 * x5^4 - 220 * x2 * x3^2 * x4 - 330 * x2 * x4^3 -
        220 * x2 * x4 * x5^2 - 200 * x2 * x3^2 - 100 * x1 * x2 * x4 +
        100 * x2^2 * x4 +
        100 * x2 * x3 * x4 +
        200 * x3^2 * x4 - 100 * x2 * x4^2 + 200 * x4^3 - 300 * x2 * x4 * x5 +
        200 * x2 * x5^2 +
        200 * x4 * x5^2 +
        121 * x2 * x4 +
        110 * x2 - 220 * x4,
        400 * x1 * x3^2 * x4^3 +
        200 * x1 * x4^5 +
        200 * x1 * x3^2 * x4 * x5^2 +
        400 * x1 * x4 * x5^4 - 220 * x1 * x3^2 * x4 - 330 * x1 * x4^3 -
        220 * x1 * x4 * x5^2 - 200 * x1 * x3^2 - 100 * x1 * x2 * x4 - 100 * x2^2 * x4 +
        100 * x1 * x3 * x4 +
        100 * x3^2 * x4 - 100 * x1 * x4^2 + 200 * x4^3 - 300 * x1 * x4 * x5 +
        200 * x1 * x5^2 +
        100 * x4 * x5^2 +
        121 * x1 * x4 +
        110 * x1 - 110 * x4 - 100,
        400 * x2^2 * x3 * x4^3 +
        200 * x3 * x4^5 +
        200 * x2^2 * x3 * x4 * x5^2 +
        400 * x3 * x4 * x5^4 - 220 * x2^2 * x3 * x4 - 330 * x3 * x4^3 -
        220 * x3 * x4 * x5^2 - 200 * x2^2 * x3 + 200 * x2^2 * x4 - 100 * x1 * x3 * x4 +
        100 * x2 * x3 * x4 +
        100 * x3^2 * x4 - 100 * x3 * x4^2 + 200 * x4^3 - 300 * x3 * x4 * x5 +
        200 * x3 * x5^2 +
        200 * x4 * x5^2 +
        121 * x3 * x4 +
        110 * x3 - 220 * x4,
        30 * x1 * x2 * x3 * x4^3 + 10 * x1 * x2 * x3 * x4 * x5^2 - 11 * x1 * x2 * x3 * x4 - 20 * x1 * x2 * x3 +
        10 * x1 * x2 * x4 +
        10 * x1 * x3 * x4 +
        10 * x2 * x3 * x4,
        12000 * x3^2 * x4^2 * x5^3 + 6000 * x3^2 * x5^5 + 6000 * x4^2 * x5^5 -
        6600 * x3^2 * x4^2 * x5 - 9900 * x3^2 * x5^3 - 9900 * x4^2 * x5^3 -
        2200 * x5^5 - 6000 * x3^2 * x4^2 - 3000 * x1 * x3^2 * x5 -
        3000 * x2 * x3^2 * x5 + 3000 * x3^2 * x4 * x5 - 3000 * x1 * x4^2 * x5 -
        3000 * x2 * x4^2 * x5 + 3000 * x3 * x4^2 * x5 - 3000 * x3^2 * x5^2 -
        3000 * x4^2 * x5^2 - 4000 * x1 * x5^3 - 4000 * x2 * x5^3 +
        5000 * x3 * x5^3 +
        5000 * x4 * x5^3 +
        2000 * x5^4 +
        3630 * x3^2 * x5 +
        3630 * x4^2 * x5 +
        3630 * x5^3 - 1100 * x1^2 - 1100 * x2^2 +
        2200 * x3^2 +
        2200 * x4^2 +
        3300 * x1 * x5 +
        3300 * x2 * x5 - 3300 * x3 * x5 - 3300 * x4 * x5 - 2200 * x5^2 +
        1000 * x1 +
        1000 * x2 - 2000 * x3 - 2000 * x4 - 4331 * x5 + 1210,
        600 * x2 * x3 * x4^2 * x5^3 + 600 * x2 * x3 * x5^5 - 220 * x2 * x3 * x4^2 * x5 -
        550 * x2 * x3 * x5^3 - 200 * x2 * x3^3 - 400 * x2 * x3 * x4^2 -
        100 * x1 * x2 * x3 * x5 +
        100 * x2^2 * x3 * x5 +
        100 * x2 * x3^2 * x5 +
        100 * x2 * x3 * x4 * x5 +
        200 * x2 * x4^2 * x5 +
        200 * x3 * x4^2 * x5 - 500 * x2 * x3 * x5^2 +
        400 * x2 * x5^3 +
        400 * x3 * x5^3 +
        121 * x2 * x3 * x5 +
        330 * x2 * x3 - 220 * x2 * x5 - 220 * x3 * x5 - 200 * x3 + 200 * x5,
        600 * x1 * x3 * x4^2 * x5^3 + 600 * x1 * x3 * x5^5 - 220 * x1 * x3 * x4^2 * x5 -
        550 * x1 * x3 * x5^3 - 200 * x1 * x3^3 - 400 * x1 * x3 * x4^2 -
        100 * x1 * x2 * x3 * x5 - 100 * x2^2 * x3 * x5 +
        100 * x1 * x3^2 * x5 +
        100 * x1 * x3 * x4 * x5 +
        200 * x1 * x4^2 * x5 +
        100 * x3 * x4^2 * x5 - 500 * x1 * x3 * x5^2 +
        400 * x1 * x5^3 +
        300 * x3 * x5^3 +
        121 * x1 * x3 * x5 +
        330 * x1 * x3 - 220 * x1 * x5 - 110 * x3 * x5 - 200 * x3 + 100 * x5,
        12000 * x2^2 * x4^2 * x5^3 + 6000 * x2^2 * x5^5 + 6000 * x4^2 * x5^5 -
        6600 * x2^2 * x4^2 * x5 - 9900 * x2^2 * x5^3 - 9900 * x4^2 * x5^3 -
        2200 * x5^5 - 6000 * x2^2 * x4^2 - 3000 * x2^2 * x3 * x5 +
        3000 * x1 * x3^2 * x5 +
        3000 * x2^2 * x4 * x5 +
        3000 * x2 * x4^2 * x5 - 3000 * x3 * x4^2 * x5 - 3000 * x2^2 * x5^2 -
        3000 * x4^2 * x5^2 - 1000 * x1 * x5^3 + 5000 * x2 * x5^3 - 4000 * x3 * x5^3 +
        5000 * x4 * x5^3 +
        2000 * x5^4 +
        3630 * x2^2 * x5 +
        3630 * x4^2 * x5 +
        3630 * x5^3 - 1100 * x1^2 + 2200 * x2^2 - 1100 * x3^2 + 2200 * x4^2 -
        3300 * x2 * x5 + 3300 * x3 * x5 - 3300 * x4 * x5 - 2200 * x5^2 + 1000 * x1 -
        2000 * x2 + 1000 * x3 - 2000 * x4 - 1331 * x5 + 1210,
        600 * x1 * x2 * x4^2 * x5^3 + 600 * x1 * x2 * x5^5 - 220 * x1 * x2 * x4^2 * x5 -
        550 * x1 * x2 * x5^3 + 200 * x1 * x2 * x3^2 - 200 * x1 * x2 * x4^2 -
        100 * x1 * x2 * x3 * x5 - 100 * x1 * x3^2 * x5 - 100 * x2 * x3^2 * x5 +
        100 * x1 * x2 * x4 * x5 +
        100 * x1 * x4^2 * x5 +
        100 * x2 * x4^2 * x5 - 300 * x1 * x2 * x5^2 +
        300 * x1 * x5^3 +
        300 * x2 * x5^3 +
        121 * x1 * x2 * x5 +
        110 * x1 * x2 - 110 * x1 * x5 - 110 * x2 * x5,
        600 * x2 * x3^2 * x4 * x5^3 + 600 * x2 * x4 * x5^5 - 220 * x2 * x3^2 * x4 * x5 -
        550 * x2 * x4 * x5^3 - 400 * x2 * x3^2 * x4 - 200 * x2 * x4^3 +
        200 * x2 * x3^2 * x5 - 100 * x1 * x2 * x4 * x5 +
        100 * x2^2 * x4 * x5 +
        100 * x2 * x3 * x4 * x5 +
        200 * x3^2 * x4 * x5 +
        100 * x2 * x4^2 * x5 - 500 * x2 * x4 * x5^2 +
        400 * x2 * x5^3 +
        400 * x4 * x5^3 +
        121 * x2 * x4 * x5 +
        330 * x2 * x4 - 220 * x2 * x5 - 220 * x4 * x5 - 200 * x4 + 200 * x5,
        600 * x1 * x3^2 * x4 * x5^3 + 600 * x1 * x4 * x5^5 - 220 * x1 * x3^2 * x4 * x5 -
        550 * x1 * x4 * x5^3 - 400 * x1 * x3^2 * x4 - 200 * x1 * x4^3 +
        200 * x1 * x3^2 * x5 - 100 * x1 * x2 * x4 * x5 - 100 * x2^2 * x4 * x5 +
        100 * x1 * x3 * x4 * x5 +
        100 * x3^2 * x4 * x5 +
        100 * x1 * x4^2 * x5 - 500 * x1 * x4 * x5^2 +
        400 * x1 * x5^3 +
        300 * x4 * x5^3 +
        121 * x1 * x4 * x5 +
        330 * x1 * x4 - 220 * x1 * x5 - 110 * x4 * x5 - 200 * x4 + 100 * x5,
        600 * x2^2 * x3 * x4 * x5^3 + 600 * x3 * x4 * x5^5 - 220 * x2^2 * x3 * x4 * x5 -
        550 * x3 * x4 * x5^3 - 400 * x2^2 * x3 * x4 - 200 * x3 * x4^3 +
        200 * x2^2 * x3 * x5 +
        200 * x2^2 * x4 * x5 - 100 * x1 * x3 * x4 * x5 +
        100 * x2 * x3 * x4 * x5 +
        100 * x3^2 * x4 * x5 +
        100 * x3 * x4^2 * x5 - 500 * x3 * x4 * x5^2 +
        400 * x3 * x5^3 +
        400 * x4 * x5^3 +
        121 * x3 * x4 * x5 +
        330 * x3 * x4 - 220 * x3 * x5 - 220 * x4 * x5 - 200 * x4 + 200 * x5,
        40 * x1 * x2 * x3 * x4 * x5^3 - 11 * x1 * x2 * x3 * x4 * x5 -
        30 * x1 * x2 * x3 * x4 +
        10 * x1 * x2 * x3 * x5 +
        10 * x1 * x2 * x4 * x5 +
        10 * x1 * x3 * x4 * x5 +
        10 * x2 * x3 * x4 * x5,
        6000 * x3 * x4^6 + 12000 * x3 * x4^2 * x5^4 + 6000 * x3 * x5^6 - 2200 * x3^5 -
        9900 * x3 * x4^4 - 6600 * x3 * x4^2 * x5^2 - 9900 * x3 * x5^4 -
        4000 * x1 * x3^3 - 4000 * x2 * x3^3 + 2000 * x3^4 - 3000 * x1 * x3 * x4^2 -
        3000 * x2 * x3 * x4^2 + 3000 * x3^2 * x4^2 - 7000 * x3 * x4^3 + 6000 * x4^4 -
        9000 * x3 * x4^2 * x5 - 3000 * x1 * x3 * x5^2 - 3000 * x2 * x3 * x5^2 +
        3000 * x3^2 * x5^2 +
        3000 * x3 * x4 * x5^2 +
        6000 * x4^2 * x5^2 - 7000 * x3 * x5^3 +
        6000 * x5^4 +
        3630 * x3^3 +
        3630 * x3 * x4^2 +
        3630 * x3 * x5^2 - 1100 * x1^2 - 1100 * x2^2 +
        3300 * x1 * x3 +
        3300 * x2 * x3 - 2200 * x3^2 + 6600 * x3 * x4 - 7700 * x4^2 + 6600 * x3 * x5 -
        7700 * x5^2 +
        1000 * x1 +
        1000 * x2 - 2331 * x3 - 3000 * x4 - 3000 * x5 + 1210,
        6000 * x2 * x4^6 + 12000 * x2 * x4^2 * x5^4 + 6000 * x2 * x5^6 - 2200 * x2^5 -
        9900 * x2 * x4^4 - 6600 * x2 * x4^2 * x5^2 - 9900 * x2 * x5^4 +
        2000 * x2^4 +
        4000 * x1 * x2 * x3^2 - 4000 * x2 * x3^3 +
        1000 * x1 * x2 * x4^2 +
        3000 * x2^2 * x4^2 - 3000 * x2 * x3 * x4^2 - 7000 * x2 * x4^3 + 6000 * x4^4 -
        9000 * x2 * x4^2 * x5 +
        1000 * x1 * x2 * x5^2 +
        3000 * x2^2 * x5^2 - 3000 * x2 * x3 * x5^2 +
        3000 * x2 * x4 * x5^2 +
        6000 * x4^2 * x5^2 - 7000 * x2 * x5^3 +
        6000 * x5^4 +
        3630 * x2^3 +
        3630 * x2 * x4^2 +
        3630 * x2 * x5^2 - 1100 * x1^2 - 1100 * x1 * x2 - 2200 * x2^2 + 3300 * x2 * x3 -
        1100 * x3^2 + 6600 * x2 * x4 - 7700 * x4^2 + 6600 * x2 * x5 - 7700 * x5^2 +
        1000 * x1 +
        5669 * x2 - 3000 * x3 - 3000 * x4 - 3000 * x5 + 1210,
        6000 * x1 * x4^6 + 12000 * x1 * x4^2 * x5^4 + 6000 * x1 * x5^6 - 2200 * x1^5 -
        9900 * x1 * x4^4 - 6600 * x1 * x4^2 * x5^2 - 9900 * x1 * x5^4 +
        2000 * x1^4 +
        4000 * x1 * x2 * x3^2 - 4000 * x1 * x3^3 + 1000 * x1 * x2 * x4^2 -
        3000 * x2^2 * x4^2 - 3000 * x1 * x3 * x4^2 - 3000 * x3^2 * x4^2 -
        7000 * x1 * x4^3 + 6000 * x4^4 - 9000 * x1 * x4^2 * x5 + 1000 * x1 * x2 * x5^2 -
        3000 * x2^2 * x5^2 - 3000 * x1 * x3 * x5^2 - 3000 * x3^2 * x5^2 +
        3000 * x1 * x4 * x5^2 - 7000 * x1 * x5^3 +
        6000 * x5^4 +
        3630 * x1^3 +
        3630 * x1 * x4^2 +
        3630 * x1 * x5^2 - 2200 * x1^2 - 1100 * x1 * x2 - 1100 * x2^2 + 3300 * x1 * x3 -
        1100 * x3^2 + 6600 * x1 * x4 - 4400 * x4^2 + 6600 * x1 * x5 - 4400 * x5^2 +
        5669 * x1 +
        1000 * x2 - 3000 * x3 - 6000 * x4 - 6000 * x5 + 1210,
        600 * x2 * x3 * x4^5 + 600 * x2 * x3 * x4 * x5^4 - 550 * x2 * x3 * x4^3 -
        220 * x2 * x3 * x4 * x5^2 - 200 * x2 * x3^3 - 100 * x1 * x2 * x3 * x4 +
        100 * x2^2 * x3 * x4 +
        100 * x2 * x3^2 * x4 - 500 * x2 * x3 * x4^2 +
        400 * x2 * x4^3 +
        400 * x3 * x4^3 - 500 * x2 * x3 * x4 * x5 +
        200 * x2 * x3 * x5^2 +
        200 * x2 * x4 * x5^2 +
        200 * x3 * x4 * x5^2 +
        121 * x2 * x3 * x4 +
        330 * x2 * x3 - 220 * x2 * x4 - 220 * x3 * x4 - 200 * x3 + 200 * x4,
        600 * x1 * x3 * x4^5 + 600 * x1 * x3 * x4 * x5^4 - 550 * x1 * x3 * x4^3 -
        220 * x1 * x3 * x4 * x5^2 - 200 * x1 * x3^3 - 100 * x1 * x2 * x3 * x4 -
        100 * x2^2 * x3 * x4 + 100 * x1 * x3^2 * x4 - 500 * x1 * x3 * x4^2 +
        400 * x1 * x4^3 +
        300 * x3 * x4^3 - 500 * x1 * x3 * x4 * x5 +
        200 * x1 * x3 * x5^2 +
        200 * x1 * x4 * x5^2 +
        100 * x3 * x4 * x5^2 +
        121 * x1 * x3 * x4 +
        330 * x1 * x3 - 220 * x1 * x4 - 110 * x3 * x4 - 200 * x3 + 100 * x4,
        600 * x1 * x2 * x4^5 + 600 * x1 * x2 * x4 * x5^4 - 550 * x1 * x2 * x4^3 -
        220 * x1 * x2 * x4 * x5^2 + 200 * x1 * x2 * x3^2 - 100 * x1 * x2 * x3 * x4 -
        100 * x1 * x3^2 * x4 - 100 * x2 * x3^2 * x4 - 300 * x1 * x2 * x4^2 +
        300 * x1 * x4^3 +
        300 * x2 * x4^3 - 500 * x1 * x2 * x4 * x5 +
        400 * x1 * x2 * x5^2 +
        100 * x1 * x4 * x5^2 +
        100 * x2 * x4 * x5^2 +
        121 * x1 * x2 * x4 +
        110 * x1 * x2 - 110 * x1 * x4 - 110 * x2 * x4,
        24000 * x4^2 * x5^6 + 2200 * x1^6 + 2200 * x2^6 + 2200 * x3^6 - 6600 * x4^6 -
        39600 * x4^2 * x5^4 - 6600 * x5^6 - 2000 * x1^5 - 2000 * x2^5 - 2000 * x3^5 -
        16000 * x1 * x4^4 - 16000 * x2 * x4^4 - 16000 * x3 * x4^4 + 6000 * x4^5 -
        12000 * x1 * x4^2 * x5^2 - 12000 * x2 * x4^2 * x5^2 - 12000 * x3 * x4^2 * x5^2 -
        28000 * x4^2 * x5^3 - 16000 * x1 * x5^4 - 16000 * x2 * x5^4 -
        16000 * x3 * x5^4 +
        20000 * x4 * x5^4 +
        6000 * x5^5 - 3630 * x1^4 - 3630 * x2^4 - 3630 * x3^4 +
        10890 * x4^4 +
        14520 * x4^2 * x5^2 +
        10890 * x5^4 +
        6600 * x1^3 +
        6600 * x2^3 +
        6600 * x3^3 +
        13200 * x1 * x4^2 +
        13200 * x2 * x4^2 +
        13200 * x3 * x4^2 - 2200 * x4^3 +
        26400 * x4^2 * x5 +
        13200 * x1 * x5^2 +
        13200 * x2 * x5^2 +
        13200 * x3 * x5^2 - 13200 * x4 * x5^2 - 2200 * x5^3 - 9669 * x1^2 +
        6000 * x1 * x2 - 9669 * x2^2 +
        6000 * x1 * x3 +
        6000 * x2 * x3 - 9669 * x3^2 +
        10000 * x1 * x4 +
        10000 * x2 * x4 +
        10000 * x3 * x4 - 34993 * x4^2 +
        10000 * x1 * x5 +
        10000 * x2 * x5 +
        10000 * x3 * x5 - 14000 * x4 * x5 - 38993 * x5^2 - 3630 * x1 - 3630 * x2 -
        3630 * x3 - 3630 * x4 - 3630 * x5 + 19800,
        24000 * x3^2 * x5^6 + 2200 * x1^6 + 2200 * x2^6 - 6600 * x3^6 + 2200 * x4^6 -
        39600 * x3^2 * x5^4 - 6600 * x5^6 - 2000 * x1^5 - 2000 * x2^5 +
        6000 * x3^5 +
        16000 * x1 * x3^2 * x4^2 +
        16000 * x2 * x3^2 * x4^2 - 16000 * x3^2 * x4^3 +
        16000 * x1 * x4^4 +
        16000 * x2 * x4^4 - 2000 * x4^5 +
        4000 * x1 * x3^2 * x5^2 +
        4000 * x2 * x3^2 * x5^2 - 12000 * x3^2 * x4 * x5^2 +
        16000 * x1 * x4^2 * x5^2 +
        16000 * x2 * x4^2 * x5^2 - 28000 * x3^2 * x5^3 + 20000 * x3 * x5^4 -
        16000 * x4 * x5^4 + 6000 * x5^5 - 3630 * x1^4 - 3630 * x2^4 + 10890 * x3^4 -
        3630 * x4^4 +
        14520 * x3^2 * x5^2 +
        10890 * x5^4 - 2200 * x1^3 - 2200 * x2^3 - 4400 * x1 * x3^2 - 4400 * x2 * x3^2 -
        2200 * x3^3 + 13200 * x3^2 * x4 - 17600 * x1 * x4^2 - 17600 * x2 * x4^2 +
        6600 * x4^3 +
        26400 * x3^2 * x5 - 4400 * x1 * x5^2 - 4400 * x2 * x5^2 - 13200 * x3 * x5^2 +
        13200 * x4 * x5^2 - 2200 * x5^3 - 1669 * x1^2 - 10000 * x1 * x2 - 1669 * x2^2 +
        2000 * x1 * x3 +
        2000 * x2 * x3 +
        13007 * x3^2 - 2000 * x1 * x4 - 2000 * x2 * x4 - 6000 * x3 * x4 +
        22331 * x4^2 +
        2000 * x1 * x5 +
        2000 * x2 * x5 - 14000 * x3 * x5 + 10000 * x4 * x5 - 6993 * x5^2 +
        6050 * x1 +
        6050 * x2 - 3630 * x3 - 3630 * x4 - 3630 * x5 + 2200,
        24000 * x2^2 * x5^6 + 2200 * x1^6 - 6600 * x2^6 + 2200 * x3^6 + 2200 * x4^6 -
        39600 * x2^2 * x5^4 - 6600 * x5^6 - 2000 * x1^5 + 6000 * x2^5 - 2000 * x3^5 +
        16000 * x2^2 * x3 * x4^2 - 16000 * x1 * x3^2 * x4^2 - 16000 * x2^2 * x4^3 +
        16000 * x3 * x4^4 - 2000 * x4^5 + 4000 * x2^2 * x3 * x5^2 -
        4000 * x1 * x3^2 * x5^2 - 12000 * x2^2 * x4 * x5^2 - 4000 * x1 * x4^2 * x5^2 +
        16000 * x3 * x4^2 * x5^2 - 28000 * x2^2 * x5^3 - 4000 * x1 * x5^4 +
        20000 * x2 * x5^4 - 16000 * x4 * x5^4 + 6000 * x5^5 - 3630 * x1^4 +
        10890 * x2^4 - 3630 * x3^4 - 3630 * x4^4 +
        14520 * x2^2 * x5^2 +
        10890 * x5^4 - 2200 * x1^3 - 2200 * x2^3 - 4400 * x2^2 * x3 + 4400 * x1 * x3^2 -
        2200 * x3^3 +
        13200 * x2^2 * x4 +
        4400 * x1 * x4^2 - 17600 * x3 * x4^2 +
        6600 * x4^3 +
        26400 * x2^2 * x5 +
        4400 * x1 * x5^2 - 13200 * x2 * x5^2 - 4400 * x3 * x5^2 + 13200 * x4 * x5^2 -
        2200 * x5^3 - 1669 * x1^2 +
        2000 * x1 * x2 +
        13007 * x2^2 - 10000 * x1 * x3 + 2000 * x2 * x3 - 1669 * x3^2 - 2000 * x1 * x4 -
        6000 * x2 * x4 - 2000 * x3 * x4 +
        6331 * x4^2 +
        2000 * x1 * x5 - 14000 * x2 * x5 +
        2000 * x3 * x5 +
        10000 * x4 * x5 - 10993 * x5^2 + 1210 * x1 - 3630 * x2 + 6050 * x3 - 3630 * x4 -
        3630 * x5 + 6600,
        18000 * x3 * x4^2 * x5^5 + 6000 * x3 * x5^7 - 16500 * x3 * x4^2 * x5^3 -
        12100 * x3 * x5^5 - 6000 * x3 * x4^4 - 3000 * x1 * x3 * x4^2 * x5 -
        3000 * x2 * x3 * x4^2 * x5 + 3000 * x3^2 * x4^2 * x5 -
        15000 * x3 * x4^2 * x5^2 - 7000 * x1 * x3 * x5^3 - 7000 * x2 * x3 * x5^3 +
        5000 * x3^2 * x5^3 +
        8000 * x3 * x4 * x5^3 +
        12000 * x4^2 * x5^3 - 7000 * x3 * x5^4 +
        6000 * x5^5 +
        3630 * x3 * x4^2 * x5 +
        7260 * x3 * x5^3 +
        2200 * x3^3 +
        9900 * x3 * x4^2 +
        3300 * x1 * x3 * x5 +
        3300 * x2 * x3 * x5 - 3300 * x3^2 * x5 - 3300 * x3 * x4 * x5 -
        6600 * x4^2 * x5 + 8800 * x3 * x5^2 - 9900 * x5^3 +
        4000 * x1 * x3 +
        4000 * x2 * x3 - 2000 * x3^2 + 1000 * x3 * x4 - 6000 * x4^2 - 3000 * x1 * x5 -
        3000 * x2 * x5 - 1331 * x3 * x5 + 3000 * x4 * x5 - 3000 * x5^2 - 3630 * x3 +
        3630 * x5 +
        1100,
        18000 * x2 * x4^2 * x5^5 + 6000 * x2 * x5^7 - 16500 * x2 * x4^2 * x5^3 -
        12100 * x2 * x5^5 - 6000 * x2 * x4^4 - 3000 * x1 * x2 * x4^2 * x5 +
        3000 * x2^2 * x4^2 * x5 - 3000 * x2 * x3 * x4^2 * x5 -
        15000 * x2 * x4^2 * x5^2 - 7000 * x1 * x2 * x5^3 + 5000 * x2^2 * x5^3 -
        7000 * x2 * x3 * x5^3 +
        8000 * x2 * x4 * x5^3 +
        12000 * x4^2 * x5^3 - 7000 * x2 * x5^4 +
        6000 * x5^5 +
        3630 * x2 * x4^2 * x5 +
        7260 * x2 * x5^3 +
        2200 * x2^3 +
        9900 * x2 * x4^2 +
        3300 * x1 * x2 * x5 - 3300 * x2^2 * x5 + 3300 * x2 * x3 * x5 -
        3300 * x2 * x4 * x5 - 6600 * x4^2 * x5 + 8800 * x2 * x5^2 - 9900 * x5^3 +
        4000 * x1 * x2 - 2000 * x2^2 +
        4000 * x2 * x3 +
        1000 * x2 * x4 - 6000 * x4^2 - 3000 * x1 * x5 - 1331 * x2 * x5 -
        3000 * x3 * x5 + 3000 * x4 * x5 - 3000 * x5^2 - 3630 * x2 +
        3630 * x5 +
        1100,
        18000 * x1 * x4^2 * x5^5 + 6000 * x1 * x5^7 - 16500 * x1 * x4^2 * x5^3 -
        12100 * x1 * x5^5 - 6000 * x1 * x4^4 - 3000 * x1 * x2 * x4^2 * x5 -
        3000 * x2^2 * x4^2 * x5 - 3000 * x1 * x3 * x4^2 * x5 - 3000 * x3^2 * x4^2 * x5 -
        15000 * x1 * x4^2 * x5^2 - 7000 * x1 * x2 * x5^3 - 5000 * x2^2 * x5^3 -
        7000 * x1 * x3 * x5^3 - 5000 * x3^2 * x5^3 +
        8000 * x1 * x4 * x5^3 +
        4000 * x4^2 * x5^3 - 7000 * x1 * x5^4 +
        6000 * x5^5 +
        3630 * x1 * x4^2 * x5 +
        7260 * x1 * x5^3 +
        2200 * x1^3 +
        9900 * x1 * x4^2 +
        3300 * x1 * x2 * x5 +
        3300 * x2^2 * x5 +
        3300 * x1 * x3 * x5 +
        3300 * x3^2 * x5 - 3300 * x1 * x4 * x5 + 8800 * x1 * x5^2 - 4400 * x5^3 -
        2000 * x1^2 +
        4000 * x1 * x2 +
        4000 * x1 * x3 +
        1000 * x1 * x4 - 6000 * x4^2 - 1331 * x1 * x5 - 3000 * x2 * x5 -
        3000 * x3 * x5 - 8000 * x5^2 - 3630 * x1 + 4400,
        18000 * x3^2 * x4 * x5^5 + 6000 * x4 * x5^7 - 16500 * x3^2 * x4 * x5^3 -
        12100 * x4 * x5^5 - 6000 * x3^2 * x4^3 - 3000 * x1 * x3^2 * x4 * x5 -
        3000 * x2 * x3^2 * x4 * x5 + 3000 * x3^2 * x4^2 * x5 -
        15000 * x3^2 * x4 * x5^2 + 12000 * x3^2 * x5^3 - 7000 * x1 * x4 * x5^3 -
        7000 * x2 * x4 * x5^3 +
        8000 * x3 * x4 * x5^3 +
        5000 * x4^2 * x5^3 - 7000 * x4 * x5^4 +
        6000 * x5^5 +
        3630 * x3^2 * x4 * x5 +
        7260 * x4 * x5^3 +
        9900 * x3^2 * x4 +
        2200 * x4^3 - 6600 * x3^2 * x5 +
        3300 * x1 * x4 * x5 +
        3300 * x2 * x4 * x5 - 3300 * x3 * x4 * x5 - 3300 * x4^2 * x5 +
        8800 * x4 * x5^2 - 9900 * x5^3 +
        4000 * x1 * x4 +
        4000 * x2 * x4 - 5000 * x3 * x4 - 2000 * x4^2 - 3000 * x1 * x5 -
        3000 * x2 * x5 + 3000 * x3 * x5 - 1331 * x4 * x5 - 3000 * x5^2 - 3630 * x4 +
        3630 * x5 +
        1100,
        1200 * x2 * x3 * x4 * x5^5 - 770 * x2 * x3 * x4 * x5^3 - 600 * x2 * x3 * x4^3 -
        100 * x1 * x2 * x3 * x4 * x5 +
        100 * x2^2 * x3 * x4 * x5 +
        100 * x2 * x3^2 * x4 * x5 +
        100 * x2 * x3 * x4^2 * x5 - 1100 * x2 * x3 * x4 * x5^2 +
        600 * x2 * x3 * x5^3 +
        600 * x2 * x4 * x5^3 +
        600 * x3 * x4 * x5^3 +
        121 * x2 * x3 * x4 * x5 +
        550 * x2 * x3 * x4 - 220 * x2 * x3 * x5 - 220 * x2 * x4 * x5 -
        220 * x3 * x4 * x5 + 200 * x2 * x3 - 400 * x2 * x4 - 400 * x3 * x4 +
        200 * x2 * x5 +
        200 * x3 * x5 +
        200 * x4 * x5,
        1200 * x1 * x3 * x4 * x5^5 - 770 * x1 * x3 * x4 * x5^3 - 600 * x1 * x3 * x4^3 -
        100 * x1 * x2 * x3 * x4 * x5 - 100 * x2^2 * x3 * x4 * x5 +
        100 * x1 * x3^2 * x4 * x5 +
        100 * x1 * x3 * x4^2 * x5 - 1100 * x1 * x3 * x4 * x5^2 +
        600 * x1 * x3 * x5^3 +
        600 * x1 * x4 * x5^3 +
        400 * x3 * x4 * x5^3 +
        121 * x1 * x3 * x4 * x5 +
        550 * x1 * x3 * x4 - 220 * x1 * x3 * x5 - 220 * x1 * x4 * x5 -
        110 * x3 * x4 * x5 + 200 * x1 * x3 - 400 * x1 * x4 - 300 * x3 * x4 +
        200 * x1 * x5 +
        100 * x3 * x5 +
        100 * x4 * x5,
        18000 * x2^2 * x4 * x5^5 + 6000 * x4 * x5^7 - 16500 * x2^2 * x4 * x5^3 -
        12100 * x4 * x5^5 - 6000 * x2^2 * x4^3 - 3000 * x2^2 * x3 * x4 * x5 +
        3000 * x1 * x3^2 * x4 * x5 +
        3000 * x2^2 * x4^2 * x5 - 15000 * x2^2 * x4 * x5^2 + 12000 * x2^2 * x5^3 -
        1000 * x1 * x4 * x5^3 + 8000 * x2 * x4 * x5^3 - 7000 * x3 * x4 * x5^3 +
        5000 * x4^2 * x5^3 - 7000 * x4 * x5^4 +
        6000 * x5^5 +
        3630 * x2^2 * x4 * x5 +
        7260 * x4 * x5^3 +
        9900 * x2^2 * x4 +
        2200 * x4^3 - 6600 * x2^2 * x5 - 3300 * x2 * x4 * x5 + 3300 * x3 * x4 * x5 -
        3300 * x4^2 * x5 + 8800 * x4 * x5^2 - 9900 * x5^3 + 1000 * x1 * x4 -
        5000 * x2 * x4 + 4000 * x3 * x4 - 2000 * x4^2 + 3000 * x2 * x5 -
        3000 * x3 * x5 + 1669 * x4 * x5 - 3000 * x5^2 - 3630 * x4 +
        3630 * x5 +
        1100,
        1200 * x1 * x2 * x4 * x5^5 - 770 * x1 * x2 * x4 * x5^3 - 600 * x1 * x2 * x4^3 -
        100 * x1 * x2 * x3 * x4 * x5 - 100 * x1 * x3^2 * x4 * x5 -
        100 * x2 * x3^2 * x4 * x5 + 100 * x1 * x2 * x4^2 * x5 -
        1100 * x1 * x2 * x4 * x5^2 +
        600 * x1 * x2 * x5^3 +
        400 * x1 * x4 * x5^3 +
        400 * x2 * x4 * x5^3 +
        121 * x1 * x2 * x4 * x5 +
        550 * x1 * x2 * x4 - 220 * x1 * x2 * x5 - 110 * x1 * x4 * x5 -
        110 * x2 * x4 * x5 + 200 * x1 * x2 - 300 * x1 * x4 - 300 * x2 * x4 +
        100 * x1 * x5 +
        100 * x2 * x5,
        18000 * x2 * x3^2 * x5^5 + 6000 * x2 * x5^7 - 16500 * x2 * x3^2 * x5^3 -
        12100 * x2 * x5^5 +
        6000 * x2 * x3^2 * x4^2 +
        6000 * x2 * x4^4 - 3000 * x2 * x3^2 * x4 * x5 + 3000 * x1 * x2 * x4^2 * x5 -
        3000 * x2^2 * x4^2 * x5 - 3000 * x3^2 * x4^2 * x5 - 9000 * x2 * x3^2 * x5^2 +
        6000 * x2 * x4^2 * x5^2 - 1000 * x1 * x2 * x5^3 +
        2000 * x2^2 * x5^3 +
        8000 * x2 * x3 * x5^3 +
        9000 * x3^2 * x5^3 - 7000 * x2 * x4 * x5^3 - 3000 * x4^2 * x5^3 -
        1000 * x2 * x5^4 +
        6000 * x5^5 +
        3630 * x2 * x3^2 * x5 +
        7260 * x2 * x5^3 - 1100 * x2^3 + 3300 * x2 * x3^2 - 6600 * x2 * x4^2 -
        3300 * x2 * x3 * x5 - 3300 * x3^2 * x5 +
        3300 * x2 * x4 * x5 +
        3300 * x4^2 * x5 +
        2200 * x2 * x5^2 - 8250 * x5^3 - 2000 * x1 * x2 + 1000 * x2^2 - 2000 * x2 * x3 +
        1000 * x2 * x4 +
        6000 * x4^2 +
        1500 * x1 * x5 - 2831 * x2 * x5 + 1500 * x3 * x5 - 4500 * x4 * x5 +
        1500 * x5^2 +
        1815 * x5 - 550,
        18000 * x1 * x3^2 * x5^5 + 6000 * x1 * x5^7 - 16500 * x1 * x3^2 * x5^3 -
        12100 * x1 * x5^5 +
        6000 * x1 * x3^2 * x4^2 +
        6000 * x1 * x4^4 - 3000 * x1 * x3^2 * x4 * x5 +
        3000 * x1 * x2 * x4^2 * x5 +
        3000 * x2^2 * x4^2 * x5 - 9000 * x1 * x3^2 * x5^2 + 6000 * x1 * x4^2 * x5^2 -
        1000 * x1 * x2 * x5^3 - 2000 * x2^2 * x5^3 +
        8000 * x1 * x3 * x5^3 +
        7000 * x3^2 * x5^3 - 7000 * x1 * x4 * x5^3 - 2000 * x4^2 * x5^3 -
        1000 * x1 * x5^4 +
        6000 * x5^5 +
        3630 * x1 * x3^2 * x5 +
        7260 * x1 * x5^3 - 1100 * x1^3 + 3300 * x1 * x3^2 - 6600 * x1 * x4^2 -
        3300 * x1 * x3 * x5 - 3300 * x3^2 * x5 +
        3300 * x1 * x4 * x5 +
        2200 * x1 * x5^2 - 6050 * x5^3 + 1000 * x1^2 - 2000 * x1 * x2 - 2000 * x1 * x3 +
        1000 * x1 * x4 +
        6000 * x4^2 - 2831 * x1 * x5 +
        1500 * x2 * x5 +
        1500 * x3 * x5 - 1500 * x4 * x5 - 500 * x5^2 + 1815 * x5 - 550,
        18000 * x2^2 * x3 * x5^5 + 6000 * x3 * x5^7 - 16500 * x2^2 * x3 * x5^3 -
        12100 * x3 * x5^5 +
        6000 * x2^2 * x3 * x4^2 +
        6000 * x3 * x4^4 - 3000 * x2^2 * x3 * x4 * x5 - 3000 * x2^2 * x4^2 * x5 +
        3000 * x1 * x3 * x4^2 * x5 - 3000 * x3^2 * x4^2 * x5 - 9000 * x2^2 * x3 * x5^2 +
        6000 * x3 * x4^2 * x5^2 +
        9000 * x2^2 * x5^3 - 1000 * x1 * x3 * x5^3 +
        8000 * x2 * x3 * x5^3 +
        2000 * x3^2 * x5^3 - 7000 * x3 * x4 * x5^3 - 3000 * x4^2 * x5^3 -
        1000 * x3 * x5^4 +
        6000 * x5^5 +
        3630 * x2^2 * x3 * x5 +
        7260 * x3 * x5^3 +
        3300 * x2^2 * x3 - 1100 * x3^3 - 6600 * x3 * x4^2 - 3300 * x2^2 * x5 -
        3300 * x2 * x3 * x5 +
        3300 * x3 * x4 * x5 +
        3300 * x4^2 * x5 +
        2200 * x3 * x5^2 - 8250 * x5^3 - 2000 * x1 * x3 - 2000 * x2 * x3 +
        1000 * x3^2 +
        1000 * x3 * x4 +
        6000 * x4^2 +
        1500 * x1 * x5 +
        1500 * x2 * x5 - 2831 * x3 * x5 - 4500 * x4 * x5 +
        1500 * x5^2 +
        1815 * x5 - 550,
        1200 * x1 * x2 * x3 * x5^5 - 770 * x1 * x2 * x3 * x5^3 + 300 * x1 * x2 * x3 * x4^2 - 100 * x1 * x2 * x3 * x4 * x5 -
        100 * x1 * x2 * x4^2 * x5 - 100 * x1 * x3 * x4^2 * x5 -
        100 * x2 * x3 * x4^2 * x5 - 800 * x1 * x2 * x3 * x5^2 +
        400 * x1 * x2 * x5^3 +
        400 * x1 * x3 * x5^3 +
        400 * x2 * x3 * x5^3 +
        121 * x1 * x2 * x3 * x5 +
        220 * x1 * x2 * x3 - 110 * x1 * x2 * x5 - 110 * x1 * x3 * x5 -
        110 * x2 * x3 * x5,
        66000 * x4^8 - 66000 * x5^8 - 60000 * x4^7 - 360000 * x1 * x4^2 * x5^4 -
        360000 * x2 * x4^2 * x5^4 - 360000 * x3 * x4^2 * x5^4 + 180000 * x4^2 * x5^5 -
        360000 * x1 * x5^6 - 360000 * x2 * x5^6 - 360000 * x3 * x5^6 -
        180000 * x4 * x5^6 + 60000 * x5^7 - 133100 * x4^6 +
        133100 * x5^6 +
        66000 * x1^5 +
        66000 * x2^5 +
        66000 * x3^5 +
        55000 * x1 * x4^4 +
        55000 * x2 * x4^4 +
        55000 * x3 * x4^4 +
        44000 * x4^5 +
        198000 * x1 * x4^2 * x5^2 +
        198000 * x2 * x4^2 * x5^2 +
        198000 * x3 * x4^2 * x5^2 - 242000 * x4^2 * x5^3 +
        539000 * x1 * x5^4 +
        539000 * x2 * x5^4 +
        539000 * x3 * x5^4 +
        242000 * x4 * x5^4 - 44000 * x5^5 - 60000 * x1^4 - 60000 * x2^4 -
        240000 * x1 * x2 * x3^2 +
        240000 * x1 * x3^3 +
        240000 * x2 * x3^3 - 60000 * x3^4 - 120000 * x1 * x2 * x4^2 +
        120000 * x1 * x3 * x4^2 +
        120000 * x2 * x3 * x4^2 +
        70000 * x1 * x4^3 +
        70000 * x2 * x4^3 +
        70000 * x3 * x4^3 - 90140 * x4^4 +
        210000 * x1 * x4^2 * x5 +
        210000 * x2 * x4^2 * x5 +
        210000 * x3 * x4^2 * x5 +
        240000 * x1 * x3 * x5^2 +
        240000 * x2 * x3 * x5^2 - 30000 * x1 * x4 * x5^2 - 30000 * x2 * x4 * x5^2 -
        30000 * x3 * x4 * x5^2 - 540000 * x4^2 * x5^2 +
        350000 * x1 * x5^3 +
        350000 * x2 * x5^3 +
        350000 * x3 * x5^3 +
        360000 * x4 * x5^3 - 1169860 * x5^4 - 108900 * x1^3 - 108900 * x2^3 -
        108900 * x3^3 - 36300 * x1 * x4^2 - 36300 * x2 * x4^2 - 36300 * x3 * x4^2 +
        48400 * x4^3 +
        72600 * x4^2 * x5 - 181500 * x1 * x5^2 - 181500 * x2 * x5^2 -
        181500 * x3 * x5^2 - 72600 * x4 * x5^2 - 48400 * x5^3 +
        132000 * x1^2 +
        66000 * x1 * x2 +
        132000 * x2^2 - 198000 * x1 * x3 - 198000 * x2 * x3 + 132000 * x3^2 -
        88000 * x1 * x4 - 88000 * x2 * x4 - 88000 * x3 * x4 + 436359 * x4^2 -
        308000 * x1 * x5 - 308000 * x2 * x5 - 308000 * x3 * x5 - 242000 * x4 * x5 +
        993641 * x5^2 - 230070 * x1 - 230070 * x2 +
        249930 * x3 +
        63380 * x4 +
        656620 * x5 - 108900,
        66000 * x3^8 - 66000 * x5^8 - 60000 * x3^7 - 360000 * x1 * x3^2 * x5^4 -
        360000 * x2 * x3^2 * x5^4 - 360000 * x3^2 * x4 * x5^4 + 180000 * x3^2 * x5^5 -
        360000 * x1 * x5^6 - 360000 * x2 * x5^6 - 180000 * x3 * x5^6 -
        360000 * x4 * x5^6 + 60000 * x5^7 - 133100 * x3^6 +
        133100 * x5^6 +
        66000 * x1^5 +
        66000 * x2^5 +
        44000 * x3^5 - 55000 * x1 * x3^2 * x4^2 - 55000 * x2 * x3^2 * x4^2 +
        55000 * x3^2 * x4^3 - 55000 * x1 * x4^4 - 55000 * x2 * x4^4 +
        66000 * x4^5 +
        143000 * x1 * x3^2 * x5^2 +
        143000 * x2 * x3^2 * x5^2 +
        198000 * x3^2 * x4 * x5^2 - 55000 * x1 * x4^2 * x5^2 -
        55000 * x2 * x4^2 * x5^2 - 242000 * x3^2 * x5^3 +
        484000 * x1 * x5^4 +
        484000 * x2 * x5^4 +
        242000 * x3 * x5^4 +
        539000 * x4 * x5^4 - 44000 * x5^5 - 60000 * x1^4 - 60000 * x2^4 -
        120000 * x1 * x2 * x3^2 +
        70000 * x1 * x3^3 +
        70000 * x2 * x3^3 - 90140 * x3^4 +
        120000 * x1 * x3^2 * x4 +
        120000 * x2 * x3^2 * x4 - 240000 * x1 * x2 * x4^2 +
        240000 * x1 * x4^3 +
        240000 * x2 * x4^3 +
        70000 * x3 * x4^3 - 60000 * x4^4 +
        210000 * x1 * x3^2 * x5 +
        210000 * x2 * x3^2 * x5 +
        210000 * x3^2 * x4 * x5 - 30000 * x1 * x3 * x5^2 - 30000 * x2 * x3 * x5^2 -
        540000 * x3^2 * x5^2 +
        240000 * x1 * x4 * x5^2 +
        240000 * x2 * x4 * x5^2 - 30000 * x3 * x4 * x5^2 +
        350000 * x1 * x5^3 +
        350000 * x2 * x5^3 +
        360000 * x3 * x5^3 +
        350000 * x4 * x5^3 - 1169860 * x5^4 - 78650 * x1^3 - 78650 * x2^3 +
        24200 * x1 * x3^2 +
        24200 * x2 * x3^2 +
        48400 * x3^3 - 36300 * x3^2 * x4 +
        60500 * x1 * x4^2 +
        60500 * x2 * x4^2 - 108900 * x4^3 + 72600 * x3^2 * x5 - 121000 * x1 * x5^2 -
        121000 * x2 * x5^2 - 72600 * x3 * x5^2 - 181500 * x4 * x5^2 - 48400 * x5^3 +
        104500 * x1^2 +
        121000 * x1 * x2 +
        104500 * x2^2 - 60500 * x1 * x3 - 60500 * x2 * x3 + 271359 * x3^2 -
        170500 * x1 * x4 - 170500 * x2 * x4 - 33000 * x3 * x4 + 22000 * x4^2 -
        280500 * x1 * x5 - 280500 * x2 * x5 - 242000 * x3 * x5 - 308000 * x4 * x5 +
        883641 * x5^2 - 263345 * x1 - 263345 * x2 - 6620 * x3 +
        319930 * x4 +
        656620 * x5 - 48400,
        66000 * x2^8 - 66000 * x5^8 - 60000 * x2^7 - 360000 * x2^2 * x3 * x5^4 +
        360000 * x1 * x3^2 * x5^4 - 360000 * x2^2 * x4 * x5^4 +
        360000 * x1 * x4^2 * x5^4 +
        180000 * x2^2 * x5^5 - 180000 * x2 * x5^6 - 360000 * x3 * x5^6 -
        360000 * x4 * x5^6 + 60000 * x5^7 - 133100 * x2^6 +
        133100 * x5^6 +
        66000 * x1^5 +
        44000 * x2^5 +
        66000 * x3^5 - 55000 * x2^2 * x3 * x4^2 +
        55000 * x1 * x3^2 * x4^2 +
        55000 * x2^2 * x4^3 - 55000 * x3 * x4^4 +
        66000 * x4^5 +
        143000 * x2^2 * x3 * x5^2 - 143000 * x1 * x3^2 * x5^2 +
        198000 * x2^2 * x4 * x5^2 - 143000 * x1 * x4^2 * x5^2 -
        55000 * x3 * x4^2 * x5^2 - 242000 * x2^2 * x5^3 - 55000 * x1 * x5^4 +
        242000 * x2 * x5^4 +
        484000 * x3 * x5^4 +
        539000 * x4 * x5^4 - 44000 * x5^5 - 60000 * x1^4 - 90140 * x2^4 -
        70000 * x1 * x2 * x3^2 +
        120000 * x1 * x3^3 +
        70000 * x2 * x3^3 - 60000 * x3^4 + 120000 * x2^2 * x3 * x4 -
        120000 * x1 * x3^2 * x4 - 70000 * x1 * x2 * x4^2 - 120000 * x1 * x3 * x4^2 +
        120000 * x1 * x4^3 +
        70000 * x2 * x4^3 +
        240000 * x3 * x4^3 - 60000 * x4^4 + 210000 * x2^2 * x3 * x5 -
        210000 * x1 * x3^2 * x5 + 210000 * x2^2 * x4 * x5 - 210000 * x1 * x4^2 * x5 -
        100000 * x1 * x2 * x5^2 - 540000 * x2^2 * x5^2 + 120000 * x1 * x3 * x5^2 -
        30000 * x2 * x3 * x5^2 + 120000 * x1 * x4 * x5^2 - 30000 * x2 * x4 * x5^2 +
        240000 * x3 * x4 * x5^2 +
        140000 * x1 * x5^3 +
        360000 * x2 * x5^3 +
        350000 * x3 * x5^3 +
        350000 * x4 * x5^3 - 809860 * x5^4 - 78650 * x1^3 +
        48400 * x2^3 +
        24200 * x2^2 * x3 - 24200 * x1 * x3^2 - 78650 * x3^3 - 36300 * x2^2 * x4 -
        24200 * x1 * x4^2 + 60500 * x3 * x4^2 - 108900 * x4^3 +
        72600 * x2^2 * x5 +
        12100 * x1 * x5^2 - 72600 * x2 * x5^2 - 121000 * x3 * x5^2 -
        181500 * x4 * x5^2 - 48400 * x5^3 +
        104500 * x1^2 +
        16500 * x1 * x2 +
        271359 * x2^2 - 11000 * x1 * x3 - 60500 * x2 * x3 + 104500 * x3^2 -
        38500 * x1 * x4 - 33000 * x2 * x4 - 170500 * x3 * x4 + 77000 * x4^2 -
        49500 * x1 * x5 - 242000 * x2 * x5 - 280500 * x3 * x5 - 308000 * x4 * x5 +
        740641 * x5^2 - 236725 * x1 - 146620 * x2 - 73345 * x3 +
        199930 * x4 +
        446620 * x5 - 72600,
        66000 * x1^8 - 66000 * x5^8 - 60000 * x1^7 +
        360000 * x2^2 * x3 * x5^4 +
        360000 * x2 * x3^2 * x5^4 +
        360000 * x2^2 * x4 * x5^4 +
        360000 * x3^2 * x4 * x5^4 +
        360000 * x2 * x4^2 * x5^4 +
        360000 * x3 * x4^2 * x5^4 - 180000 * x2^2 * x5^5 - 180000 * x3^2 * x5^5 -
        180000 * x4^2 * x5^5 - 180000 * x1 * x5^6 + 60000 * x5^7 - 133100 * x1^6 +
        133100 * x5^6 +
        44000 * x1^5 +
        66000 * x2^5 +
        66000 * x3^5 +
        55000 * x2^2 * x3 * x4^2 +
        55000 * x2 * x3^2 * x4^2 - 55000 * x2^2 * x4^3 - 55000 * x3^2 * x4^3 +
        66000 * x4^5 - 143000 * x2^2 * x3 * x5^2 - 143000 * x2 * x3^2 * x5^2 -
        198000 * x2^2 * x4 * x5^2 - 198000 * x3^2 * x4 * x5^2 -
        143000 * x2 * x4^2 * x5^2 - 143000 * x3 * x4^2 * x5^2 +
        242000 * x2^2 * x5^3 +
        242000 * x3^2 * x5^3 +
        242000 * x4^2 * x5^3 +
        242000 * x1 * x5^4 - 55000 * x2 * x5^4 - 55000 * x3 * x5^4 -
        110000 * x4 * x5^4 + 154000 * x5^5 - 90140 * x1^4 - 60000 * x2^4 -
        70000 * x1 * x2 * x3^2 +
        70000 * x1 * x3^3 +
        120000 * x2 * x3^3 - 60000 * x3^4 - 120000 * x2^2 * x3 * x4 -
        120000 * x2 * x3^2 * x4 - 70000 * x1 * x2 * x4^2 - 120000 * x2 * x3 * x4^2 +
        70000 * x1 * x4^3 +
        120000 * x2 * x4^3 +
        120000 * x3 * x4^3 - 60000 * x4^4 - 210000 * x2^2 * x3 * x5 -
        210000 * x2 * x3^2 * x5 - 210000 * x2^2 * x4 * x5 - 210000 * x3^2 * x4 * x5 -
        210000 * x2 * x4^2 * x5 - 210000 * x3 * x4^2 * x5 - 100000 * x1 * x2 * x5^2 +
        540000 * x2^2 * x5^2 - 30000 * x1 * x3 * x5^2 +
        120000 * x2 * x3 * x5^2 +
        540000 * x3^2 * x5^2 - 30000 * x1 * x4 * x5^2 +
        120000 * x2 * x4 * x5^2 +
        120000 * x3 * x4 * x5^2 +
        540000 * x4^2 * x5^2 +
        360000 * x1 * x5^3 +
        140000 * x2 * x5^3 +
        140000 * x3 * x5^3 +
        140000 * x4 * x5^3 - 269860 * x5^4 + 48400 * x1^3 - 78650 * x2^3 -
        24200 * x2^2 * x3 - 24200 * x2 * x3^2 - 78650 * x3^3 +
        36300 * x2^2 * x4 +
        36300 * x3^2 * x4 - 24200 * x2 * x4^2 - 24200 * x3 * x4^2 - 48400 * x4^3 -
        72600 * x2^2 * x5 - 72600 * x3^2 * x5 - 72600 * x4^2 * x5 - 72600 * x1 * x5^2 +
        12100 * x2 * x5^2 +
        12100 * x3 * x5^2 +
        72600 * x4 * x5^2 - 314600 * x5^3 +
        271359 * x1^2 +
        16500 * x1 * x2 +
        104500 * x2^2 - 60500 * x1 * x3 - 11000 * x2 * x3 + 104500 * x3^2 -
        33000 * x1 * x4 - 38500 * x2 * x4 - 38500 * x3 * x4 + 77000 * x4^2 -
        242000 * x1 * x5 - 49500 * x2 * x5 - 49500 * x3 * x5 - 22000 * x4 * x5 -
        7359 * x5^2 - 146620 * x1 - 236725 * x2 - 46725 * x3 +
        40000 * x4 +
        646480 * x5 - 133100,
        2640000 * x5^9 +
        9600000 * x1 * x5^7 +
        9600000 * x2 * x5^7 +
        9600000 * x3 * x5^7 +
        9600000 * x4 * x5^7 - 2400000 * x5^8 - 6050000 * x5^7 +
        660000 * x1^6 +
        660000 * x2^6 +
        660000 * x3^6 +
        660000 * x4^6 - 16500000 * x1 * x5^5 - 16500000 * x2 * x5^5 -
        16500000 * x3 * x5^5 - 16500000 * x4 * x5^5 + 440000 * x5^6 - 600000 * x1^5 -
        600000 * x2^5 - 600000 * x3^5 - 600000 * x4^5 - 8000000 * x1 * x2 * x5^3 -
        8000000 * x1 * x3 * x5^3 - 8000000 * x2 * x3 * x5^3 - 8000000 * x1 * x4 * x5^3 -
        8000000 * x2 * x4 * x5^3 - 8000000 * x3 * x4 * x5^3 - 13200000 * x1 * x5^4 -
        13200000 * x2 * x5^4 - 13200000 * x3 * x5^4 - 13200000 * x4 * x5^4 +
        38058500 * x5^5 - 1331000 * x1^4 - 1331000 * x2^4 - 1331000 * x3^4 -
        1331000 * x4^4 +
        8470000 * x1 * x5^3 +
        8470000 * x2 * x5^3 +
        8470000 * x3 * x5^3 +
        8470000 * x4 * x5^3 +
        4719000 * x5^4 +
        3080000 * x1^3 +
        3080000 * x2^3 +
        3080000 * x3^3 +
        3080000 * x4^3 +
        3300000 * x1 * x2 * x5 +
        3300000 * x1 * x3 * x5 +
        3300000 * x2 * x3 * x5 +
        3300000 * x1 * x4 * x5 +
        3300000 * x2 * x4 * x5 +
        3300000 * x3 * x4 * x5 +
        13200000 * x1 * x5^2 +
        13200000 * x2 * x5^2 +
        13200000 * x3 * x5^2 +
        13200000 * x4 * x5^2 - 38094100 * x5^3 - 3301400 * x1^2 + 5000000 * x1 * x2 -
        3301400 * x2^2 +
        5000000 * x1 * x3 +
        5000000 * x2 * x3 - 3301400 * x3^2 +
        5000000 * x1 * x4 +
        5000000 * x2 * x4 +
        5000000 * x3 * x4 - 3301400 * x4^2 - 2331000 * x1 * x5 - 2331000 * x2 * x5 -
        2331000 * x3 * x5 - 2331000 * x4 * x5 - 31828900 * x5^2 - 3509000 * x1 -
        3509000 * x2 - 3509000 * x3 - 3509000 * x4 +
        8147051 * x5 +
        9863590,
        240000 * x4 * x5^8 - 66000 * x4^7 - 484000 * x4 * x5^6 - 180000 * x1 * x4^5 -
        180000 * x2 * x4^5 - 180000 * x3 * x4^5 + 60000 * x4^6 -
        280000 * x1 * x4 * x5^4 - 280000 * x2 * x4 * x5^4 - 280000 * x3 * x4 * x5^4 +
        200000 * x4^2 * x5^4 - 460000 * x4 * x5^5 +
        240000 * x5^6 +
        133100 * x4^5 +
        290400 * x4 * x5^4 - 22000 * x1^4 - 22000 * x2^4 - 22000 * x3^4 +
        242000 * x1 * x4^3 +
        242000 * x2 * x4^3 +
        242000 * x3 * x4^3 - 44000 * x4^4 +
        132000 * x1 * x4 * x5^2 +
        132000 * x2 * x4 * x5^2 +
        132000 * x3 * x4 * x5^2 - 132000 * x4^2 * x5^2 + 682000 * x4 * x5^3 -
        418000 * x5^4 +
        20000 * x1^3 +
        20000 * x2^3 +
        20000 * x3^3 +
        60000 * x1 * x2 * x4 +
        60000 * x1 * x3 * x4 +
        60000 * x2 * x3 * x4 +
        120000 * x1 * x4^2 +
        120000 * x2 * x4^2 +
        120000 * x3 * x4^2 - 449860 * x4^3 +
        220000 * x1 * x4 * x5 +
        220000 * x2 * x4 * x5 +
        220000 * x3 * x4 * x5 - 140000 * x4^2 * x5 - 120000 * x1 * x5^2 -
        120000 * x2 * x5^2 - 120000 * x3 * x5^2 + 6760 * x4 * x5^2 - 280000 * x5^3 +
        36300 * x1^2 +
        36300 * x2^2 +
        36300 * x3^2 - 72600 * x1 * x4 - 72600 * x2 * x4 - 72600 * x3 * x4 -
        24200 * x4^2 - 217800 * x4 * x5 + 181500 * x5^2 - 88000 * x1 - 88000 * x2 -
        88000 * x3 +
        69641 * x4 +
        286000 * x5 +
        116690,
        240000 * x3 * x5^8 - 66000 * x3^7 - 484000 * x3 * x5^6 +
        60000 * x3^6 +
        360000 * x1 * x3 * x4^4 +
        360000 * x2 * x3 * x4^4 - 180000 * x3 * x4^5 +
        180000 * x1 * x3 * x4^2 * x5^2 +
        180000 * x2 * x3 * x4^2 * x5^2 +
        80000 * x1 * x3 * x5^4 +
        80000 * x2 * x3 * x5^4 +
        200000 * x3^2 * x5^4 - 280000 * x3 * x4 * x5^4 - 460000 * x3 * x5^5 +
        240000 * x5^6 +
        133100 * x3^5 +
        290400 * x3 * x5^4 - 22000 * x1^4 - 22000 * x2^4 - 55000 * x1 * x3^3 -
        55000 * x2 * x3^3 - 44000 * x3^4 - 198000 * x1 * x3 * x4^2 -
        198000 * x2 * x3 * x4^2 + 242000 * x3 * x4^3 - 22000 * x4^4 -
        66000 * x1 * x3 * x5^2 - 66000 * x2 * x3 * x5^2 - 132000 * x3^2 * x5^2 +
        132000 * x3 * x4 * x5^2 +
        682000 * x3 * x5^3 - 418000 * x5^4 +
        20000 * x1^3 +
        20000 * x2^3 - 120000 * x1 * x2 * x3 +
        30000 * x1 * x3^2 +
        30000 * x2 * x3^2 +
        90140 * x3^3 - 210000 * x1 * x3 * x4 - 210000 * x2 * x3 * x4 -
        60000 * x3^2 * x4 +
        180000 * x1 * x4^2 +
        180000 * x2 * x4^2 +
        450000 * x3 * x4^2 - 160000 * x4^3 - 50000 * x1 * x3 * x5 -
        50000 * x2 * x3 * x5 - 140000 * x3^2 * x5 +
        220000 * x3 * x4 * x5 +
        60000 * x1 * x5^2 +
        60000 * x2 * x5^2 +
        276760 * x3 * x5^2 - 120000 * x4 * x5^2 - 280000 * x5^3 +
        36300 * x1^2 +
        36300 * x2^2 +
        36300 * x1 * x3 +
        36300 * x2 * x3 - 24200 * x3^2 - 72600 * x3 * x4 + 36300 * x4^2 -
        217800 * x3 * x5 +
        181500 * x5^2 +
        11000 * x1 +
        11000 * x2 - 469359 * x3 +
        154000 * x4 +
        286000 * x5 +
        26690,
        240000 * x2 * x5^8 - 66000 * x2^7 - 484000 * x2 * x5^6 +
        60000 * x2^6 +
        360000 * x1 * x2 * x4^4 +
        360000 * x2 * x3 * x4^4 - 180000 * x2 * x4^5 +
        180000 * x1 * x2 * x4^2 * x5^2 +
        180000 * x2 * x3 * x4^2 * x5^2 +
        80000 * x1 * x2 * x5^4 +
        200000 * x2^2 * x5^4 +
        80000 * x2 * x3 * x5^4 - 280000 * x2 * x4 * x5^4 - 460000 * x2 * x5^5 +
        240000 * x5^6 +
        133100 * x2^5 +
        290400 * x2 * x5^4 - 22000 * x1^4 - 44000 * x2^4 + 55000 * x1 * x2 * x3^2 -
        55000 * x2 * x3^3 - 22000 * x3^4 - 143000 * x1 * x2 * x4^2 -
        198000 * x2 * x3 * x4^2 + 242000 * x2 * x4^3 - 22000 * x4^4 -
        11000 * x1 * x2 * x5^2 - 132000 * x2^2 * x5^2 - 66000 * x2 * x3 * x5^2 +
        132000 * x2 * x4 * x5^2 +
        682000 * x2 * x5^3 - 418000 * x5^4 +
        20000 * x1^3 +
        90140 * x2^3 - 120000 * x1 * x2 * x3 + 30000 * x2^2 * x3 - 30000 * x1 * x3^2 +
        20000 * x3^3 - 210000 * x1 * x2 * x4 - 60000 * x2^2 * x4 -
        210000 * x2 * x3 * x4 +
        150000 * x1 * x4^2 +
        450000 * x2 * x4^2 +
        180000 * x3 * x4^2 - 160000 * x4^3 - 50000 * x1 * x2 * x5 - 140000 * x2^2 * x5 -
        50000 * x2 * x3 * x5 +
        220000 * x2 * x4 * x5 +
        30000 * x1 * x5^2 +
        276760 * x2 * x5^2 +
        60000 * x3 * x5^2 - 120000 * x4 * x5^2 - 280000 * x5^3 + 36300 * x1^2 -
        24200 * x1 * x2 - 24200 * x2^2 +
        36300 * x2 * x3 +
        36300 * x3^2 - 72600 * x2 * x4 + 36300 * x4^2 - 217800 * x2 * x5 +
        181500 * x5^2 +
        44000 * x1 - 359359 * x2 - 44000 * x3 +
        154000 * x4 +
        286000 * x5 - 3310,
        240000 * x1 * x5^8 - 66000 * x1^7 - 484000 * x1 * x5^6 +
        60000 * x1^6 +
        360000 * x1 * x2 * x4^4 +
        360000 * x1 * x3 * x4^4 - 180000 * x1 * x4^5 +
        180000 * x1 * x2 * x4^2 * x5^2 +
        180000 * x1 * x3 * x4^2 * x5^2 +
        80000 * x1 * x2 * x5^4 - 200000 * x2^2 * x5^4 + 80000 * x1 * x3 * x5^4 -
        200000 * x3^2 * x5^4 - 280000 * x1 * x4 * x5^4 - 200000 * x4^2 * x5^4 -
        460000 * x1 * x5^5 +
        240000 * x5^6 +
        133100 * x1^5 +
        290400 * x1 * x5^4 - 44000 * x1^4 - 22000 * x2^4 + 55000 * x1 * x2 * x3^2 -
        55000 * x1 * x3^3 - 22000 * x3^4 - 143000 * x1 * x2 * x4^2 -
        198000 * x1 * x3 * x4^2 + 242000 * x1 * x4^3 - 22000 * x4^4 -
        11000 * x1 * x2 * x5^2 + 132000 * x2^2 * x5^2 - 66000 * x1 * x3 * x5^2 +
        132000 * x3^2 * x5^2 +
        132000 * x1 * x4 * x5^2 +
        132000 * x4^2 * x5^2 +
        682000 * x1 * x5^3 - 198000 * x5^4 +
        90140 * x1^3 +
        20000 * x2^3 - 120000 * x1 * x2 * x3 - 30000 * x2^2 * x3 - 30000 * x2 * x3^2 +
        20000 * x3^3 - 210000 * x1 * x2 * x4 + 60000 * x2^2 * x4 -
        210000 * x1 * x3 * x4 +
        60000 * x3^2 * x4 +
        450000 * x1 * x4^2 +
        150000 * x2 * x4^2 +
        150000 * x3 * x4^2 - 160000 * x4^3 - 50000 * x1 * x2 * x5 + 140000 * x2^2 * x5 -
        50000 * x1 * x3 * x5 +
        140000 * x3^2 * x5 +
        220000 * x1 * x4 * x5 +
        140000 * x4^2 * x5 +
        276760 * x1 * x5^2 +
        30000 * x2 * x5^2 +
        30000 * x3 * x5^2 - 60000 * x4 * x5^2 - 480000 * x5^3 - 24200 * x1^2 -
        24200 * x1 * x2 +
        36300 * x2^2 +
        36300 * x1 * x3 +
        36300 * x3^2 - 72600 * x1 * x4 + 36300 * x4^2 - 217800 * x1 * x5 +
        36300 * x5^2 - 359359 * x1 + 44000 * x2 - 11000 * x3 +
        88000 * x4 +
        264000 * x5 +
        166690,
        24000 * x3 * x4 * x5^7 - 28600 * x3 * x4 * x5^5 - 6000 * x3 * x4^5 -
        10000 * x1 * x3 * x4 * x5^3 - 10000 * x2 * x3 * x4 * x5^3 +
        8000 * x3^2 * x4 * x5^3 +
        8000 * x3 * x4^2 * x5^3 - 40000 * x3 * x4 * x5^4 +
        18000 * x3 * x5^5 +
        18000 * x4 * x5^5 +
        10890 * x3 * x4 * x5^3 +
        12100 * x3 * x4^3 +
        3300 * x1 * x3 * x4 * x5 +
        3300 * x2 * x3 * x4 * x5 - 3300 * x3^2 * x4 * x5 - 3300 * x3 * x4^2 * x5 +
        25300 * x3 * x4 * x5^2 - 16500 * x3 * x5^3 - 16500 * x4 * x5^3 +
        7000 * x1 * x3 * x4 +
        7000 * x2 * x3 * x4 - 5000 * x3^2 * x4 + 1000 * x3 * x4^2 - 6000 * x4^3 -
        3000 * x1 * x3 * x5 - 3000 * x2 * x3 * x5 + 3000 * x3^2 * x5 -
        3000 * x1 * x4 * x5 - 3000 * x2 * x4 * x5 +
        13669 * x3 * x4 * x5 +
        3000 * x4^2 * x5 - 15000 * x3 * x5^2 - 15000 * x4 * x5^2 + 12000 * x5^3 -
        7260 * x3 * x4 +
        3630 * x3 * x5 +
        3630 * x4 * x5 - 2200 * x3 + 9900 * x4 - 6600 * x5,
        24000 * x2 * x4 * x5^7 - 28600 * x2 * x4 * x5^5 - 6000 * x2 * x4^5 -
        10000 * x1 * x2 * x4 * x5^3 + 8000 * x2^2 * x4 * x5^3 -
        10000 * x2 * x3 * x4 * x5^3 + 8000 * x2 * x4^2 * x5^3 - 40000 * x2 * x4 * x5^4 +
        18000 * x2 * x5^5 +
        18000 * x4 * x5^5 +
        10890 * x2 * x4 * x5^3 +
        12100 * x2 * x4^3 +
        3300 * x1 * x2 * x4 * x5 - 3300 * x2^2 * x4 * x5 + 3300 * x2 * x3 * x4 * x5 -
        3300 * x2 * x4^2 * x5 + 25300 * x2 * x4 * x5^2 - 16500 * x2 * x5^3 -
        16500 * x4 * x5^3 + 7000 * x1 * x2 * x4 - 5000 * x2^2 * x4 +
        7000 * x2 * x3 * x4 +
        1000 * x2 * x4^2 - 6000 * x4^3 - 3000 * x1 * x2 * x5 + 3000 * x2^2 * x5 -
        3000 * x2 * x3 * x5 - 3000 * x1 * x4 * x5 + 13669 * x2 * x4 * x5 -
        3000 * x3 * x4 * x5 + 3000 * x4^2 * x5 - 15000 * x2 * x5^2 - 15000 * x4 * x5^2 +
        12000 * x5^3 - 7260 * x2 * x4 +
        3630 * x2 * x5 +
        3630 * x4 * x5 - 2200 * x2 + 9900 * x4 - 6600 * x5,
        24000 * x1 * x4 * x5^7 - 28600 * x1 * x4 * x5^5 - 6000 * x1 * x4^5 -
        10000 * x1 * x2 * x4 * x5^3 - 8000 * x2^2 * x4 * x5^3 -
        10000 * x1 * x3 * x4 * x5^3 - 8000 * x3^2 * x4 * x5^3 +
        8000 * x1 * x4^2 * x5^3 - 40000 * x1 * x4 * x5^4 +
        18000 * x1 * x5^5 +
        10000 * x4 * x5^5 +
        10890 * x1 * x4 * x5^3 +
        12100 * x1 * x4^3 +
        3300 * x1 * x2 * x4 * x5 +
        3300 * x2^2 * x4 * x5 +
        3300 * x1 * x3 * x4 * x5 +
        3300 * x3^2 * x4 * x5 - 3300 * x1 * x4^2 * x5 + 25300 * x1 * x4 * x5^2 -
        16500 * x1 * x5^3 - 4400 * x4 * x5^3 +
        7000 * x1 * x2 * x4 +
        5000 * x2^2 * x4 +
        7000 * x1 * x3 * x4 +
        5000 * x3^2 * x4 +
        1000 * x1 * x4^2 - 6000 * x4^3 - 3000 * x1 * x2 * x5 - 3000 * x2^2 * x5 -
        3000 * x1 * x3 * x5 - 3000 * x3^2 * x5 + 13669 * x1 * x4 * x5 -
        3000 * x2 * x4 * x5 - 3000 * x3 * x4 * x5 - 15000 * x1 * x5^2 -
        10000 * x4 * x5^2 + 4000 * x5^3 - 7260 * x1 * x4 + 3630 * x1 * x5 - 2200 * x1 +
        4400 * x4 +
        2000,
        24000 * x2 * x3 * x5^7 - 28600 * x2 * x3 * x5^5 +
        12000 * x2 * x3 * x4^4 +
        6000 * x2 * x3 * x4^2 * x5^2 - 10000 * x1 * x2 * x3 * x5^3 +
        8000 * x2^2 * x3 * x5^3 +
        8000 * x2 * x3^2 * x5^3 - 10000 * x2 * x3 * x4 * x5^3 - 28000 * x2 * x3 * x5^4 +
        18000 * x2 * x5^5 +
        18000 * x3 * x5^5 +
        10890 * x2 * x3 * x5^3 +
        2200 * x2 * x3^3 - 6600 * x2 * x3 * x4^2 + 3300 * x1 * x2 * x3 * x5 -
        3300 * x2^2 * x3 * x5 - 3300 * x2 * x3^2 * x5 +
        3300 * x2 * x3 * x4 * x5 +
        18700 * x2 * x3 * x5^2 - 16500 * x2 * x5^3 - 16500 * x3 * x5^3 +
        4000 * x1 * x2 * x3 - 2000 * x2^2 * x3 - 2000 * x2 * x3^2 -
        2000 * x2 * x3 * x4 +
        6000 * x2 * x4^2 +
        6000 * x3 * x4^2 - 3000 * x1 * x2 * x5 + 3000 * x2^2 * x5 -
        3000 * x1 * x3 * x5 +
        4669 * x2 * x3 * x5 +
        3000 * x3^2 * x5 - 3000 * x2 * x4 * x5 - 3000 * x3 * x4 * x5 -
        9000 * x2 * x5^2 - 9000 * x3 * x5^2 + 12000 * x5^3 - 3630 * x2 * x3 +
        3630 * x2 * x5 +
        3630 * x3 * x5 +
        1100 * x2 +
        3300 * x3 - 6600 * x5,
        24000 * x1 * x3 * x5^7 - 28600 * x1 * x3 * x5^5 +
        12000 * x1 * x3 * x4^4 +
        6000 * x1 * x3 * x4^2 * x5^2 - 10000 * x1 * x2 * x3 * x5^3 -
        8000 * x2^2 * x3 * x5^3 + 8000 * x1 * x3^2 * x5^3 -
        10000 * x1 * x3 * x4 * x5^3 - 8000 * x3 * x4^2 * x5^3 - 28000 * x1 * x3 * x5^4 +
        18000 * x1 * x5^5 +
        10000 * x3 * x5^5 +
        10890 * x1 * x3 * x5^3 +
        2200 * x1 * x3^3 - 6600 * x1 * x3 * x4^2 +
        3300 * x1 * x2 * x3 * x5 +
        3300 * x2^2 * x3 * x5 - 3300 * x1 * x3^2 * x5 +
        3300 * x1 * x3 * x4 * x5 +
        3300 * x3 * x4^2 * x5 +
        18700 * x1 * x3 * x5^2 - 16500 * x1 * x5^3 - 4400 * x3 * x5^3 +
        4000 * x1 * x2 * x3 +
        2000 * x2^2 * x3 - 2000 * x1 * x3^2 - 2000 * x1 * x3 * x4 +
        6000 * x1 * x4^2 +
        8000 * x3 * x4^2 - 3000 * x1 * x2 * x5 - 3000 * x2^2 * x5 +
        4669 * x1 * x3 * x5 - 3000 * x2 * x3 * x5 - 3000 * x1 * x4 * x5 -
        3000 * x3 * x4 * x5 - 3000 * x4^2 * x5 - 9000 * x1 * x5^2 - 7000 * x3 * x5^2 +
        4000 * x5^3 - 3630 * x1 * x3 +
        3630 * x1 * x5 +
        1100 * x1 +
        1100 * x3 - 1000,
        24000 * x1 * x2 * x5^7 - 28600 * x1 * x2 * x5^5 +
        12000 * x1 * x2 * x4^4 +
        6000 * x1 * x2 * x4^2 * x5^2 - 10000 * x1 * x2 * x3 * x5^3 -
        8000 * x1 * x3^2 * x5^3 - 8000 * x2 * x3^2 * x5^3 -
        10000 * x1 * x2 * x4 * x5^3 - 8000 * x1 * x4^2 * x5^3 -
        8000 * x2 * x4^2 * x5^3 - 28000 * x1 * x2 * x5^4 +
        10000 * x1 * x5^5 +
        10000 * x2 * x5^5 +
        10890 * x1 * x2 * x5^3 - 2200 * x1 * x2 * x3^2 - 8800 * x1 * x2 * x4^2 +
        3300 * x1 * x2 * x3 * x5 +
        3300 * x1 * x3^2 * x5 +
        3300 * x2 * x3^2 * x5 +
        3300 * x1 * x2 * x4 * x5 +
        3300 * x1 * x4^2 * x5 +
        3300 * x2 * x4^2 * x5 +
        16500 * x1 * x2 * x5^2 - 4400 * x1 * x5^3 - 4400 * x2 * x5^3 +
        4000 * x1 * x2 * x3 +
        2000 * x1 * x3^2 +
        2000 * x2 * x3^2 - 2000 * x1 * x2 * x4 +
        8000 * x1 * x4^2 +
        8000 * x2 * x4^2 +
        4669 * x1 * x2 * x5 - 3000 * x1 * x3 * x5 - 3000 * x2 * x3 * x5 -
        3000 * x3^2 * x5 - 3000 * x1 * x4 * x5 - 3000 * x2 * x4 * x5 -
        3000 * x4^2 * x5 - 7000 * x1 * x5^2 - 7000 * x2 * x5^2 - 4000 * x5^3 -
        1210 * x1 * x2 - 1100 * x1 - 1100 * x2 +
        3300 * x5 +
        1000
    ]

    @test gb == map(f -> divexact(f, leading_coefficient(f)), sing_ans)

    kat = Groebner.katsuran(6, ordering=:degrevlex)
    (x1, x2, x3, x4, x5, x6, x7) = gens(parent(first(kat)))

    gb = Groebner.groebner(kat)

    sing_ans = [
        x1 + 2 * x2 + 2 * x3 + 2 * x4 + 2 * x5 + 2 * x6 + 2 * x7 - 1,
        26 * x4 * x5 +
        28 * x5^2 +
        26 * x3 * x6 +
        56 * x4 * x6 +
        90 * x5 * x6 +
        64 * x6^2 +
        26 * x2 * x7 +
        56 * x3 * x7 +
        90 * x4 * x7 +
        128 * x5 * x7 +
        170 * x6 * x7 +
        108 * x7^2 - x2 - 4 * x3 - 9 * x4 - 16 * x5 - 25 * x6 - 36 * x7,
        13 * x4^2 + 26 * x3 * x5 - 15 * x5^2 + 26 * x2 * x6 - 30 * x4 * x6 - 64 * x5 * x6 - 51 * x6^2 - 52 * x2 * x7 - 82 * x3 * x7 - 116 * x4 * x7 -
        154 * x5 * x7 - 196 * x6 * x7 - 147 * x7^2 +
        x2 +
        4 * x3 +
        9 * x4 +
        16 * x5 +
        25 * x6 +
        49 * x7,
        2 * x3 * x4 + 2 * x2 * x5 - 4 * x2 * x6 - 4 * x3 * x6 - 4 * x4 * x6 - 4 * x5 * x6 - 4 * x6^2 + 2 * x2 * x7 - 4 * x6 * x7 + x6,
        13 * x3^2 + 26 * x2 * x4 - 52 * x2 * x5 - 52 * x3 * x5 +
        4 * x5^2 +
        26 * x2 * x6 +
        52 * x3 * x6 +
        112 * x4 * x6 +
        128 * x5 * x6 +
        128 * x6^2 +
        52 * x2 * x7 +
        138 * x3 * x7 +
        180 * x4 * x7 +
        204 * x5 * x7 +
        340 * x6 * x7 +
        216 * x7^2 - 2 * x2 - 8 * x3 - 18 * x4 - 19 * x5 - 50 * x6 - 72 * x7,
        26 * x2 * x3 - 52 * x2 * x4 + 78 * x2 * x5 + 104 * x3 * x5 - 4 * x5^2 -
        26 * x3 * x6 - 164 * x4 * x6 - 180 * x5 * x6 - 180 * x6^2 - 104 * x2 * x7 -
        216 * x3 * x7 - 310 * x4 * x7 - 360 * x5 * x7 - 548 * x6 * x7 - 372 * x7^2 +
        2 * x2 +
        8 * x3 +
        31 * x4 +
        32 * x5 +
        76 * x6 +
        124 * x7,
        13 * x2^2 + 26 * x2 * x4 - 26 * x3 * x5 +
        8 * x5^2 +
        42 * x4 * x6 +
        48 * x5 * x6 +
        48 * x6^2 +
        52 * x2 * x7 +
        68 * x3 * x7 +
        100 * x4 * x7 +
        122 * x5 * x7 +
        160 * x6 * x7 +
        120 * x7^2 - 4 * x2 - 3 * x3 - 10 * x4 - 12 * x5 - 22 * x6 - 40 * x7,
        270 * x5^2 * x7 +
        540 * x4 * x6 * x7 +
        1152 * x5 * x6 * x7 +
        918 * x6^2 * x7 +
        540 * x3 * x7^2 +
        1152 * x4 * x7^2 +
        1836 * x5 * x7^2 +
        2592 * x6 * x7^2 +
        1710 * x7^3 +
        x5^2 +
        2 * x4 * x6 +
        6 * x5 * x6 +
        6 * x6^2 - 18 * x2 * x7 - 70 * x3 * x7 - 156 * x4 * x7 - 276 * x5 * x7 -
        430 * x6 * x7 - 633 * x7^2 +
        6 * x2 +
        11 * x3 +
        15 * x4 +
        18 * x5 +
        20 * x6 +
        21 * x7,
        64800 * x5^2 * x6 +
        129600 * x4 * x6^2 +
        276480 * x5 * x6^2 +
        220320 * x6^3 +
        129600 * x3 * x6 * x7 +
        276480 * x4 * x6 * x7 +
        591804 * x5 * x6 * x7 +
        782136 * x6^2 * x7 +
        151164 * x4 * x7^2 +
        320112 * x5 * x7^2 +
        917244 * x6 * x7^2 +
        355680 * x7^3 +
        532 * x5^2 - 4320 * x2 * x6 - 17280 * x3 * x6 - 37816 * x4 * x6 -
        65928 * x5 * x6 - 104808 * x6^2 - 4446 * x2 * x7 - 16720 * x3 * x7 -
        36822 * x4 * x7 - 64752 * x5 * x7 - 256030 * x6 * x7 - 152076 * x7^2 +
        1437 * x2 +
        5852 * x3 +
        7980 * x4 +
        9576 * x5 +
        10640 * x6 +
        11172 * x7,
        1053000 * x3 * x5 * x6 +
        842400 * x2 * x6^2 +
        898560 * x3 * x6^2 +
        795600 * x4 * x6^2 +
        861120 * x5 * x6^2 +
        608400 * x6^3 +
        631800 * x2 * x5 * x7 +
        1797120 * x3 * x5 * x7 +
        168480 * x2 * x6 * x7 +
        1172340 * x3 * x6 * x7 +
        313560 * x4 * x6 * x7 +
        1140516 * x5 * x6 * x7 +
        1370304 * x6^2 * x7 - 985140 * x2 * x7^2 - 393120 * x3 * x7^2 +
        57096 * x4 * x7^2 +
        971568 * x5 * x7^2 +
        1112436 * x6 * x7^2 +
        411840 * x7^3 +
        14040 * x2 * x4 - 33930 * x2 * x5 - 163800 * x3 * x5 + 39568 * x5^2 -
        128700 * x2 * x6 - 265590 * x3 * x6 - 63604 * x4 * x6 - 19992 * x5 * x6 -
        172092 * x6^2 + 5616 * x2 * x7 - 202600 * x3 * x7 - 34968 * x4 * x7 -
        151368 * x5 * x7 - 271000 * x6 * x7 - 187104 * x7^2 +
        8673 * x2 +
        21068 * x3 +
        14370 * x4 +
        24264 * x5 +
        46655 * x6 +
        16608 * x7,
        1170000 * x2 * x5 * x6 - 530400 * x2 * x6^2 - 926640 * x3 * x6^2 -
        2152800 * x4 * x6^2 - 2695680 * x5 * x6^2 - 2371200 * x6^3 +
        1170000 * x2 * x4 * x7 - 1060800 * x2 * x5 * x7 - 1853280 * x3 * x5 * x7 +
        1541280 * x2 * x6 * x7 - 1007760 * x3 * x6 * x7 + 165360 * x4 * x6 * x7 -
        1196364 * x5 * x6 * x7 - 3897816 * x6^2 * x7 +
        1756560 * x2 * x7^2 +
        2976480 * x3 * x7^2 +
        3106116 * x4 * x7^2 +
        2712528 * x5 * x7^2 +
        1755156 * x6 * x7^2 +
        2845440 * x7^3 - 95160 * x2 * x4 - 20280 * x2 * x5 + 109200 * x3 * x5 -
        27172 * x5^2 - 93600 * x2 * x6 +
        188760 * x3 * x6 +
        228016 * x4 * x6 +
        258168 * x5 * x6 +
        850968 * x6^2 - 285714 * x2 * x7 - 136400 * x3 * x7 - 428778 * x4 * x7 -
        317328 * x5 * x7 + 244750 * x6 * x7 - 1052484 * x7^2 - 7617 * x2 - 30572 * x3 -
        17580 * x4 - 30456 * x5 - 21620 * x6 + 34668 * x7,
        327600 * x5^3 - 655200 * x3 * x6^2 - 2795520 * x4 * x6^2 - 3325920 * x5 * x6^2 -
        3138720 * x6^3 + 655200 * x3 * x5 * x7 - 655200 * x2 * x6 * x7 -
        4193280 * x3 * x6 * x7 - 8895120 * x4 * x6 * x7 - 12666108 * x5 * x6 * x7 -
        15459912 * x6^2 * x7 - 1397760 * x2 * x7^2 - 4455360 * x3 * x7^2 -
        9533628 * x4 * x7^2 - 12227904 * x5 * x7^2 - 24840348 * x6 * x7^2 -
        12673440 * x7^3 - 21840 * x2 * x5 - 87360 * x3 * x5 - 141524 * x5^2 +
        48360 * x2 * x6 +
        390000 * x3 * x6 +
        851072 * x4 * x6 +
        885576 * x5 * x6 +
        1670256 * x6^2 +
        322062 * x2 * x7 +
        917840 * x3 * x7 +
        1787334 * x4 * x7 +
        2144304 * x5 * x7 +
        6088430 * x6 * x7 +
        5278092 * x7^2 - 31509 * x2 - 91924 * x3 - 124500 * x4 - 188712 * x5 -
        264280 * x6 - 351204 * x7,
        12285000 * x3 * x5^2 - 3931200 * x2 * x6^2 - 4651920 * x3 * x6^2 -
        3494400 * x4 * x6^2 +
        2988960 * x5 * x6^2 +
        5834400 * x6^3 - 24570000 * x2 * x4 * x7 - 7862400 * x2 * x5 * x7 -
        9303840 * x3 * x5 * x7 - 39017160 * x2 * x6 * x7 - 23030280 * x3 * x6 * x7 -
        43300920 * x4 * x6 * x7 - 44663892 * x5 * x6 * x7 - 20070648 * x6^2 * x7 -
        6235320 * x2 * x7^2 - 38689560 * x3 * x7^2 - 59641452 * x4 * x7^2 -
        75774816 * x5 * x7^2 - 86197332 * x6 * x7^2 - 53707680 * x7^3 +
        1294020 * x2 * x4 +
        660660 * x2 * x5 +
        54600 * x3 * x5 - 1050916 * x5^2 +
        3365700 * x2 * x6 +
        2340780 * x3 * x6 +
        578248 * x4 * x6 - 2194896 * x5 * x6 - 3653496 * x6^2 +
        5511558 * x2 * x7 +
        7058800 * x3 * x7 +
        9850566 * x4 * x7 +
        10807416 * x5 * x7 +
        12994750 * x6 * x7 +
        21800748 * x7^2 - 145251 * x2 - 49616 * x3 +
        206760 * x4 +
        477432 * x5 - 333110 * x6 - 1299396 * x7,
        3510000 * x2 * x5^2 +
        7020000 * x2 * x4 * x6 +
        11138400 * x2 * x6^2 +
        13515840 * x3 * x6^2 +
        26176800 * x4 * x6^2 +
        26782080 * x5 * x6^2 +
        24835200 * x6^3 +
        14040000 * x2 * x4 * x7 +
        1216800 * x2 * x5 * x7 - 1048320 * x3 * x5 * x7 +
        30532320 * x2 * x6 * x7 +
        46879560 * x3 * x6 * x7 +
        79581840 * x4 * x6 * x7 +
        94464084 * x5 * x6 * x7 +
        114857496 * x6^2 * x7 +
        15629640 * x2 * x7^2 +
        42981120 * x3 * x7^2 +
        70716204 * x4 * x7^2 +
        88510032 * x5 * x7^2 +
        175658964 * x6 * x7^2 +
        84483360 * x7^3 - 1067040 * x2 * x4 - 560820 * x2 * x5 - 109200 * x3 * x5 +
        445732 * x5^2 - 3455400 * x2 * x6 - 4506060 * x3 * x6 - 6925696 * x4 * x6 -
        6365808 * x5 * x6 - 12574608 * x6^2 - 4921566 * x2 * x7 - 8233600 * x3 * x7 -
        13216782 * x4 * x7 - 14999232 * x5 * x7 - 43228750 * x6 * x7 - 35519196 * x7^2 +
        138027 * x2 +
        368132 * x3 +
        601980 * x4 +
        825336 * x5 +
        1795970 * x6 +
        2452692 * x7,
        23256 * x5 * x6 * x7^2 +
        24624 * x6^2 * x7^2 +
        23256 * x4 * x7^3 +
        49248 * x5 * x7^3 +
        77976 * x6 * x7^3 +
        54720 * x7^4 - 64 * x5 * x6^2 - 96 * x6^3 - 64 * x4 * x6 * x7 +
        36 * x5 * x6 * x7 - 80 * x6^2 * x7 - 684 * x2 * x7^2 - 2736 * x3 * x7^2 -
        5928 * x4 * x7^2 - 10336 * x5 * x7^2 - 16280 * x6 * x7^2 - 23712 * x7^3 -
        224 * x2 * x6 - 416 * x3 * x6 - 576 * x4 * x6 - 704 * x5 * x6 - 800 * x6^2 +
        224 * x2 * x7 +
        912 * x3 * x7 +
        1254 * x4 * x7 +
        1520 * x5 * x7 +
        846 * x6 * x7 +
        1824 * x7^2 - x2,
        707719615200 * x5 * x6^2 * x7 +
        749350180800 * x6^3 * x7 +
        707719615200 * x4 * x6 * x7^2 +
        1196626672800 * x6^2 * x7^2 - 1498700361600 * x4 * x7^3 -
        2352631132800 * x5 * x7^3 - 1631226729600 * x6 * x7^3 - 2165077512000 * x7^4 +
        3809332800 * x4 * x6^2 +
        12019307040 * x5 * x6^2 +
        15444056160 * x6^3 - 20815282800 * x2 * x6 * x7 - 77547132000 * x3 * x6 * x7 -
        161033240160 * x4 * x6 * x7 - 294410760228 * x5 * x6 * x7 -
        453957338952 * x6^2 * x7 +
        24376528800 * x2 * x7^2 +
        101315448000 * x3 * x7^2 +
        221420412252 * x4 * x7^2 +
        379186428816 * x5 * x7^2 - 110014529508 * x6 * x7^2 + 832832900640 * x7^3 -
        6666332400 * x2 * x5 - 12380331600 * x3 * x5 - 2461514524 * x5^2 +
        14245840440 * x2 * x6 +
        43161262560 * x3 * x6 +
        72299905312 * x4 * x6 +
        77853438696 * x5 * x6 +
        88875947856 * x6^2 +
        9191457522 * x2 * x7 +
        4544407840 * x3 * x7 +
        38769924954 * x4 * x7 +
        31851802464 * x5 * x7 +
        127684259410 * x6 * x7 +
        32318628132 * x7^2 - 538622259 * x2 - 2384020364 * x3 - 5495722260 * x4 -
        10023266232 * x5 - 15898628480 * x6 - 23121809004 * x7,
        353859807600 * x4 * x6^2 * x7 - 104076414000 * x6^3 * x7 +
        530789711400 * x3 * x6 * x7^2 +
        936687726000 * x4 * x6 * x7^2 - 545158738800 * x6^2 * x7^2 +
        176929903800 * x2 * x7^3 +
        749350180800 * x3 * x7^3 +
        801388387800 * x4 * x7^3 - 216075600000 * x5 * x7^3 -
        1267787570400 * x6 * x7^3 - 849033057600 * x7^4 - 462561840 * x3 * x6^2 -
        1043031600 * x4 * x6^2 - 919283040 * x5 * x6^2 - 852873840 * x6^3 +
        5203820700 * x2 * x5 * x7 +
        19890159120 * x3 * x5 * x7 +
        9019955880 * x2 * x6 * x7 - 9196817760 * x3 * x6 * x7 -
        78956186400 * x4 * x6 * x7 - 26907907572 * x5 * x6 * x7 +
        4014356112 * x6^2 * x7 - 25668714240 * x2 * x7^2 - 84752853120 * x3 * x7^2 -
        130458186612 * x4 * x7^2 +
        25310295504 * x5 * x7^2 +
        188575651368 * x6 * x7^2 +
        291195571680 * x7^3 +
        1618966440 * x2 * x4 +
        2345687370 * x2 * x5 +
        1655245800 * x3 * x5 +
        1300807564 * x5^2 +
        7514397540 * x2 * x6 +
        5555828070 * x3 * x6 +
        6329733548 * x4 * x6 +
        12018544224 * x5 * x6 +
        12582832884 * x6^2 - 10112604372 * x2 * x7 - 43686704560 * x3 * x7 -
        74894569284 * x4 * x7 - 75571776024 * x5 * x7 - 44534089390 * x6 * x7 -
        25824868932 * x7^2 - 277991391 * x2 +
        221154224 * x3 +
        1655865960 * x4 +
        4082172192 * x5 +
        5990511425 * x6 +
        7698894924 * x7,
        2300088749400 * x3 * x6^2 * x7 +
        749350180800 * x6^3 * x7 +
        4600177498800 * x3 * x5 * x7^2 +
        6900266248200 * x2 * x6 * x7^2 +
        8523858306600 * x3 * x6 * x7^2 +
        5713795128600 * x4 * x6 * x7^2 - 90247575600 * x6^2 * x7^2 -
        3111884778600 * x2 * x7^3 + 1540330927200 * x3 * x7^3 -
        6150916067400 * x4 * x7^3 - 10692212965200 * x5 * x7^3 -
        11683783893600 * x6 * x7^3 - 12717201463200 * x7^4 - 4625618400 * x2 * x6^2 +
        9174143160 * x3 * x6^2 +
        982566000 * x4 * x6^2 +
        12393875520 * x5 * x6^2 +
        23304106800 * x6^3 +
        67649669100 * x2 * x4 * x7 - 170569678500 * x2 * x5 * x7 -
        762224818680 * x3 * x5 * x7 - 540272229120 * x2 * x6 * x7 -
        1024372671660 * x3 * x6 * x7 - 211656114540 * x4 * x6 * x7 -
        129601233684 * x5 * x6 * x7 - 234383863896 * x6^2 * x7 +
        277578496260 * x2 * x7^2 - 261664083720 * x3 * x7^2 +
        1357700700096 * x4 * x7^2 +
        2125110675168 * x5 * x7^2 +
        2136019556736 * x6 * x7^2 +
        5389127490240 * x7^3 - 23725567710 * x2 * x4 +
        15857859420 * x2 * x5 +
        73129364100 * x3 * x5 - 13592803832 * x5^2 +
        56153056050 * x2 * x6 +
        134664809760 * x3 * x6 +
        96478089896 * x4 * x6 +
        75665916708 * x5 * x6 +
        100651723008 * x6^2 - 90997265034 * x2 * x7 - 244232285200 * x3 * x7 -
        338737831518 * x4 * x7 - 394059577668 * x5 * x7 - 318816305800 * x6 * x7 -
        488577635904 * x7^2 - 350224077 * x2 +
        5980430918 * x3 +
        9668662620 * x4 +
        3047268564 * x5 +
        12100584530 * x6 +
        35074730208 * x7,
        84336587478000 * x2 * x6^2 * x7 - 19080675900000 * x6^3 * x7 +
        168673174956000 * x2 * x5 * x7^2 - 128985369084000 * x2 * x6 * x7^2 -
        254588254713000 * x3 * x6 * x7^2 - 520833067794000 * x4 * x6 * x7^2 +
        68843726874000 * x6^2 * x7^2 +
        103052995929000 * x2 * x7^3 - 297242238384000 * x3 * x7^3 +
        139896046485000 * x4 * x7^3 +
        666048715488000 * x5 * x7^3 +
        797100247476000 * x6 * x7^3 +
        813993277968000 * x7^4 +
        764768908800 * x2 * x6^2 +
        223725743280 * x3 * x6^2 +
        108671799600 * x4 * x6^2 - 1499895764640 * x5 * x6^2 - 2529309603600 * x6^3 -
        13078936026000 * x2 * x4 * x7 - 1271852325900 * x2 * x5 * x7 +
        15399762964560 * x3 * x5 * x7 - 15000803377560 * x2 * x6 * x7 +
        13549007693520 * x3 * x6 * x7 +
        8017432001280 * x4 * x6 * x7 +
        11426546884428 * x5 * x6 * x7 +
        11541696858432 * x6^2 * x7 - 44450586573120 * x2 * x7^2 -
        38895981096960 * x3 * x7^2 - 89342344844532 * x4 * x7^2 -
        189640408413456 * x5 * x7^2 - 193935585839112 * x6 * x7^2 -
        387982125254880 * x7^3 + 3468288676320 * x2 * x4 - 1296064641690 * x2 * x5 -
        2686615097400 * x3 * x5 - 149729289556 * x5^2 - 2267685681300 * x2 * x6 -
        9757682217270 * x3 * x6 - 10605750093332 * x4 * x6 - 13295373601236 * x5 * x6 -
        16254858157836 * x6^2 +
        7889191711128 * x2 * x7 +
        25831236728800 * x3 * x7 +
        34635561909456 * x4 * x7 +
        39834625378056 * x5 * x7 +
        28203714543250 * x6 * x7 +
        45496600912668 * x7^2 +
        226497875709 * x2 +
        90285259744 * x3 - 507534447840 * x4 - 31582764288 * x5 - 1391138792135 * x6 -
        2204307793236 * x7,
        7084273348152000 * x2 * x4 * x6 * x7 +
        1570721240088000 * x6^3 * x7 +
        9167883156432000 * x2 * x4 * x7^2 - 16787941883856000 * x2 * x5 * x7^2 -
        9568369197504000 * x3 * x5 * x7^2 +
        14157722749248000 * x2 * x6 * x7^2 +
        8235670716234000 * x3 * x6 * x7^2 +
        27324429884388000 * x4 * x6 * x7^2 - 3066437597652000 * x6^2 * x7^2 -
        2796637320594000 * x2 * x7^3 + 16053578706672000 * x3 * x7^3 -
        16190022885426000 * x4 * x7^3 - 44845554452400000 * x5 * x7^3 -
        47143845452808000 * x6 * x7^3 - 48889072228464000 * x7^4 -
        492489591048000 * x2 * x4 * x6 - 604565241134400 * x2 * x6^2 -
        428214464758080 * x3 * x6^2 - 822048515870400 * x4 * x6^2 -
        700744024992960 * x5 * x6^2 - 562210943954400 * x6^3 -
        662966757180000 * x2 * x4 * x7 + 583429634260200 * x2 * x5 * x7 -
        240088405808160 * x3 * x5 * x7 - 1130361289097040 * x2 * x6 * x7 -
        1122788240257320 * x3 * x6 * x7 - 1613516605487280 * x4 * x6 * x7 -
        2579500058541588 * x5 * x6 * x7 - 3390399368625672 * x6^2 * x7 +
        2055184815102120 * x2 * x7^2 +
        4442973935736960 * x3 * x7^2 +
        7488572494819572 * x4 * x7^2 +
        12979106486579376 * x5 * x7^2 +
        10644715408059252 * x6 * x7^2 +
        24281503471349280 * x7^3 - 284216342782320 * x2 * x4 +
        140839984036440 * x2 * x5 +
        240084953222400 * x3 * x5 +
        37729741697276 * x5^2 +
        146983391848800 * x2 * x6 +
        581312585360520 * x3 * x6 +
        589898438963272 * x4 * x6 +
        832338453232056 * x5 * x6 +
        1195363504129656 * x6^2 - 448982248035438 * x2 * x7 -
        1614383931809600 * x3 * x7 - 1960261957375926 * x4 * x7 -
        2474092630679376 * x5 * x7 - 1299031190841050 * x6 * x7 -
        3117484299273828 * x7^2 - 10525225096239 * x2 - 11601099910124 * x3 +
        28428531457440 * x4 - 11874608182152 * x5 +
        95728363986160 * x6 +
        151922981773356 * x7,
        38499947066880 * x5 * x6^3 + 40764649835520 * x6^4 + 98697744924480 * x6^3 * x7 - 57749920600320 * x3 * x6 * x7^2 -
        183440924259840 * x4 * x6 * x7^2 + 105457684475520 * x6^2 * x7^2 -
        19249973533440 * x2 * x7^3 - 81529299671040 * x3 * x7^3 -
        3763403130240 * x4 * x7^3 +
        105673299114240 * x5 * x7^3 +
        226196696344320 * x6 * x7^3 +
        185847372622080 * x7^4 - 1132351384320 * x2 * x6^2 - 4380645649536 * x3 * x6^2 -
        10485736350336 * x4 * x6^2 - 18374317760448 * x5 * x6^2 -
        28435497214176 * x6^3 - 566175692160 * x2 * x5 * x7 -
        1967182993152 * x3 * x5 * x7 - 636115042368 * x2 * x6 * x7 +
        1105130980512 * x3 * x6 * x7 +
        7423811785728 * x4 * x6 * x7 +
        1862053603956 * x5 * x6 * x7 - 41729950528344 * x6^2 * x7 +
        1373428158336 * x2 * x7^2 +
        3529475524224 * x3 * x7^2 +
        753809438772 * x4 * x7^2 - 24964391760048 * x5 * x7^2 -
        52835833548108 * x6 * x7^2 - 77357856247776 * x7^3 - 520659607104 * x2 * x4 +
        785270375136 * x2 * x5 +
        1515214726272 * x3 * x5 - 38790629444 * x5^2 - 739273459392 * x2 * x6 -
        593460640656 * x3 * x6 - 1128723036808 * x4 * x6 - 1115943140568 * x5 * x6 -
        967150286904 * x6^2 - 166008056058 * x2 * x7 +
        3515520242192 * x3 * x7 +
        4108910830782 * x4 * x7 +
        6856846882608 * x5 * x7 +
        4744744181654 * x6 * x7 +
        2827834813788 * x7^2 +
        35597004135 * x2 +
        20083394516 * x3 - 36896291292 * x4 - 151058766600 * x5 +
        151820715848 * x6 +
        769469733228 * x7,
        31281206991840000 * x4 * x6^3 - 9200354997600000 * x6^4 -
        58617917893080000 * x6^3 * x7 - 93843620975520000 * x3 * x5 * x7^2 -
        125124827967360000 * x2 * x6 * x7^2 - 231848945939520000 * x3 * x6 * x7^2 -
        219817712481120000 * x4 * x6 * x7^2 - 60937453446480000 * x6^2 * x7^2 +
        22080851994240000 * x2 * x7^3 - 170701971186240000 * x3 * x7^3 +
        34008009038640000 * x4 * x7^3 +
        308678810183640000 * x5 * x7^3 +
        336559743496080000 * x6 * x7^3 +
        380299754343600000 * x7^4 - 278998799414400 * x2 * x6^2 -
        1883563839087840 * x3 * x6^2 - 6925102372468800 * x4 * x6^2 -
        2465685193870080 * x5 * x6^2 + 367792921360800 * x6^3 -
        1840070999520000 * x2 * x4 * x7 +
        1989793015891200 * x2 * x5 * x7 +
        10104176779744320 * x3 * x5 * x7 +
        7124679049999680 * x2 * x6 * x7 +
        16214345393725440 * x3 * x6 * x7 +
        11040074654600160 * x4 * x6 * x7 +
        14094001731320316 * x5 * x6 * x7 +
        33573038366452104 * x6^2 * x7 - 172668776660640 * x2 * x7^2 +
        19981719311654880 * x3 * x7^2 - 9608703217800804 * x4 * x7^2 -
        51683971880606832 * x5 * x7^2 - 22824598893582564 * x6 * x7^2 -
        142633707985323360 * x7^3 + 485403606221040 * x2 * x4 -
        748228486477680 * x2 * x5 - 1349100561322800 * x3 * x5 +
        229763399176468 * x5^2 - 3292858970247600 * x2 * x6 -
        6967536711445440 * x3 * x6 - 9383709379404904 * x4 * x6 -
        9895500289711992 * x5 * x6 - 9201858854607192 * x6^2 +
        2623624625828466 * x2 * x7 +
        9672798596513600 * x3 * x7 +
        12663078331109082 * x4 * x7 +
        11594852591992032 * x5 * x7 +
        1573937219929250 * x6 * x7 +
        5897383741336596 * x7^2 +
        90572571490173 * x2 +
        26789675658068 * x3 +
        131802906721020 * x4 +
        427913962098264 * x5 +
        253207251856280 * x6 - 202780891098492 * x7,
        172046638455120000 * x3 * x6^3 +
        56051393523840000 * x6^4 +
        114690126699720000 * x6^3 * x7 - 688186553820480000 * x2 * x5 * x7^2 -
        1275169202667360000 * x3 * x5 * x7^2 - 951316706751840000 * x2 * x6 * x7^2 -
        721365334030575000 * x3 * x6 * x7^2 +
        16764108385050000 * x4 * x6 * x7^2 +
        457308278896950000 * x6^2 * x7^2 +
        655812804672675000 * x2 * x7^3 +
        811152836961480000 * x3 * x7^3 +
        933796743410115000 * x4 * x7^3 +
        688927894799040000 * x5 * x7^3 +
        1485346655438100000 * x6 * x7^3 +
        1284241576087080000 * x7^4 +
        5060195248680000 * x2 * x4 * x6 - 2411436635524800 * x2 * x6^2 -
        40830972877262880 * x3 * x6^2 +
        923443129346400 * x4 * x6^2 +
        7850110719709440 * x5 * x6^2 - 11818714664390400 * x6^3 +
        39906539802090000 * x2 * x4 * x7 +
        51482856989502900 * x2 * x5 * x7 +
        144711019662384240 * x3 * x5 * x7 +
        164175027728900760 * x2 * x6 * x7 +
        124009286563274580 * x3 * x6 * x7 +
        88327707928476120 * x4 * x6 * x7 +
        76399601401824012 * x5 * x6 * x7 +
        8578717157717928 * x6^2 * x7 +
        48432189279247020 * x2 * x7^2 +
        147514755917936160 * x3 * x7^2 +
        28443833338215672 * x4 * x7^2 +
        145595842963146576 * x5 * x7^2 +
        8817387251804352 * x6 * x7^2 - 283943100412595520 * x7^3 -
        11196538536591720 * x2 * x4 +
        5402241916194240 * x2 * x5 +
        6149204679290400 * x3 * x5 +
        919870872833176 * x5^2 - 9329157017692200 * x2 * x6 -
        9028674290755080 * x3 * x6 - 10467426771610528 * x4 * x6 -
        7526515948415544 * x5 * x6 - 5830862792332944 * x6^2 -
        20313362996832438 * x2 * x7 - 38708535265025800 * x3 * x7 -
        37575052315462326 * x4 * x7 - 49528873780714176 * x5 * x7 -
        79948815054389500 * x6 * x7 - 85277225586543528 * x7^2 +
        606990867930861 * x2 +
        1378753130554376 * x3 +
        3250365364933890 * x4 +
        2068524519180048 * x5 +
        8299497831423710 * x6 +
        12410472423318456 * x7,
        344093276910240000 * x2 * x6^3 - 77849157672000000 * x6^4 -
        214932365607960000 * x6^3 * x7 - 688186553820480000 * x2 * x4 * x7^2 +
        1052520611725440000 * x2 * x5 * x7^2 +
        2077440158458080000 * x3 * x5 * x7^2 +
        1399834012884840000 * x2 * x6 * x7^2 +
        1901483369128980000 * x3 * x6 * x7^2 +
        1041550957689840000 * x4 * x6 * x7^2 - 809357011042320000 * x6^2 * x7^2 -
        1216056921807780000 * x2 * x7^3 - 856022260565160000 * x3 * x7^3 -
        1802972961749700000 * x4 * x7^3 - 2746384049302920000 * x5 * x7^3 -
        4251190930637040000 * x6 * x7^3 - 3924886027079160000 * x7^4 -
        53362058986080000 * x2 * x4 * x6 - 114440568419548800 * x2 * x6^2 -
        1023678030754080 * x3 * x6^2 - 10920835984593600 * x4 * x6^2 -
        10229097828792960 * x5 * x6^2 + 21213067204461600 * x6^3 -
        4255164186390000 * x2 * x4 * x7 - 70130780654547600 * x2 * x5 * x7 -
        212328546727808160 * x3 * x5 * x7 - 283497414214425840 * x2 * x6 * x7 -
        314102192829300720 * x3 * x6 * x7 - 283480546590802080 * x4 * x6 * x7 -
        300074707927323108 * x5 * x6 * x7 - 162588640463549352 * x6^2 * x7 -
        14440089947989680 * x2 * x7^2 - 271016376422257440 * x3 * x7^2 -
        21036952906894548 * x4 * x7^2 +
        9400160918964816 * x5 * x7^2 +
        69249863243620332 * x6 * x7^2 +
        1143017625211843680 * x7^3 +
        25768461051969480 * x2 * x4 - 6900614520986160 * x2 * x5 -
        13490559739893600 * x3 * x5 - 1929065230043084 * x5^2 +
        29652772243231800 * x2 * x6 +
        37120922304354720 * x3 * x6 +
        61767580949559752 * x4 * x6 +
        53538652688239896 * x5 * x6 +
        71601776478239496 * x6^2 +
        34754909012369442 * x2 * x7 +
        37008464645163200 * x3 * x7 +
        28515021961189434 * x4 * x7 +
        49055130522589584 * x5 * x7 +
        162307056834466250 * x6 * x7 +
        135567300436683252 * x7^2 - 1368983503628499 * x2 - 2445908955392884 * x3 -
        6681277804000260 * x4 - 5627320726275432 * x5 - 17868933819732640 * x6 -
        26824909351352604 * x7,
        114697758970080000 * x2 * x4 * x6^2 +
        25430724839520000 * x6^4 +
        58770008226072000 * x6^3 * x7 +
        128191612966560000 * x2 * x4 * x7^2 - 296864787922560000 * x2 * x5 * x7^2 -
        103657332972960000 * x3 * x5 * x7^2 +
        11193765247080000 * x2 * x6 * x7^2 +
        274553927053380000 * x3 * x6 * x7^2 +
        688162963166640000 * x4 * x6 * x7^2 +
        9981675244008000 * x6^2 * x7^2 +
        68619314357100000 * x2 * x7^3 +
        815186838768120000 * x3 * x7^3 +
        472126069269036000 * x4 * x7^3 - 203647600362024000 * x5 * x7^3 -
        340218473843664000 * x6 * x7^3 - 398077988948640000 * x7^4 +
        9200354997600000 * x2 * x4 * x6 +
        19134746295350400 * x2 * x6^2 +
        23732605536999840 * x3 * x6^2 +
        42050044571308800 * x4 * x6^2 +
        41541668469706080 * x5 * x6^2 +
        30043804621303200 * x6^3 +
        28099417555170000 * x2 * x4 * x7 +
        2503112537530800 * x2 * x5 * x7 - 4593464287420320 * x3 * x5 * x7 +
        81069770540650320 * x2 * x6 * x7 +
        81216431842768560 * x3 * x6 * x7 +
        124609129462449840 * x4 * x6 * x7 +
        146301032822316684 * x5 * x6 * x7 +
        149984288998933896 * x6^2 * x7 +
        55029467688302640 * x2 * x7^2 +
        46988943917217120 * x3 * x7^2 +
        91147973659611804 * x4 * x7^2 +
        249407005094078832 * x5 * x7^2 +
        371169257542413564 * x6 * x7^2 +
        331314200937027360 * x7^3 - 8511290321522040 * x2 * x4 +
        2410597764007680 * x2 * x5 +
        4677404823202800 * x3 * x5 +
        965164379484532 * x5^2 - 970338275813400 * x2 * x6 +
        6847855133377440 * x3 * x6 +
        5620024468807904 * x4 * x6 +
        9138348249193992 * x5 * x6 +
        720685645368192 * x6^2 - 19005533395398966 * x2 * x7 -
        49339671646279600 * x3 * x7 - 67016805659187582 * x4 * x7 -
        72927586560244032 * x5 * x7 - 98280147260320750 * x6 * x7 -
        91063802130103596 * x7^2 - 203738907168423 * x2 +
        402647695536932 * x3 +
        1810896744247980 * x4 +
        1348925269943736 * x5 +
        5263106019446720 * x6 +
        8285540937351492 * x7,
        45987620595696000 * x6^2 * x7^3 +
        91975241191392000 * x5 * x7^4 +
        193632086718720000 * x6 * x7^4 +
        152485268290992000 * x7^5 +
        96249867667200 * x6^4 +
        133634115576000 * x6^3 * x7 +
        278133808773600 * x4 * x6 * x7^2 +
        513162984204000 * x6^2 * x7^2 - 2420401083984000 * x2 * x7^3 -
        9305097500649600 * x3 * x7^3 - 20331119322072000 * x4 * x7^3 -
        35136647667950400 * x5 * x7^3 - 54662077833480000 * x6 * x7^3 -
        80973075268612800 * x7^4 +
        769998941337600 * x2 * x6^2 +
        1443748015008000 * x3 * x6^2 +
        2020717723752000 * x4 * x6^2 +
        2499612373084320 * x5 * x6^2 +
        2883529052333280 * x6^3 - 753013670572800 * x2 * x5 * x7 -
        1411900632324000 * x3 * x5 * x7 - 775244392603200 * x2 * x6 * x7 -
        1183143939494400 * x3 * x6 * x7 + 499588577673120 * x4 * x6 * x7 -
        14699782727604 * x5 * x6 * x7 +
        2960517019355064 * x6^2 * x7 +
        2733321783909600 * x2 * x7^2 +
        7965289291845600 * x3 * x7^2 +
        15207855522971436 * x4 * x7^2 +
        17220372973463088 * x5 * x7^2 +
        18673714214984556 * x6 * x7^2 +
        20109108589146720 * x7^3 - 1550398449600 * x2 * x5 +
        1720866092400 * x3 * x5 +
        1519504118468 * x5^2 +
        2795994227520 * x2 * x6 - 8225425372320 * x3 * x6 - 18616502897984 * x4 * x6 -
        17106034808472 * x5 * x6 - 20458417458192 * x6^2 - 100680521954454 * x2 * x7 -
        382356112843280 * x3 * x7 - 839049359619678 * x4 * x7 -
        1504412778438048 * x5 * x7 - 2394659941600070 * x6 * x7 -
        3437445433630524 * x7^2 +
        7120029113913 * x2 +
        13282268426548 * x3 +
        18424209388620 * x4 +
        22585598799624 * x5 +
        25756981351360 * x6 +
        27938357043828 * x7,
        14945976693601200000 * x4 * x6 * x7^3 +
        11956781354880960000 * x3 * x7^4 +
        12586085636716800000 * x4 * x7^4 +
        10737504308824020000 * x5 * x7^4 +
        11406140108274600000 * x6 * x7^4 +
        7944966558177480000 * x7^5 +
        9200354997600000 * x6^4 +
        32315726547000000 * x6^3 * x7 +
        157326070458960000 * x2 * x5 * x7^2 +
        587442666596760000 * x3 * x5 * x7^2 +
        181592006765130000 * x2 * x6 * x7^2 - 516170357716455000 * x3 * x6 * x7^2 -
        2664121506086430000 * x4 * x6 * x7^2 + 562523034076965000 * x6^2 * x7^2 -
        1276275274757145000 * x2 * x7^3 - 4350251000130750000 * x3 * x7^3 -
        6269058148561065000 * x4 * x7^3 - 3034425059383770000 * x5 * x7^3 -
        2472901111164510000 * x6 * x7^3 - 4469194022223060000 * x7^4 -
        47841845987520000 * x2 * x4 * x6 - 1154813387270400 * x2 * x6^2 +
        72891518859113760 * x3 * x6^2 +
        21496849350163200 * x4 * x6^2 +
        78024449220513120 * x5 * x6^2 +
        144966292807096800 * x6^3 - 18035401782060000 * x2 * x4 * x7 -
        270289350311748300 * x2 * x5 * x7 - 602151443660337480 * x3 * x5 * x7 -
        258634134182567520 * x2 * x6 * x7 - 141370518857370660 * x3 * x6 * x7 +
        410051516045216760 * x4 * x6 * x7 +
        348142503530159676 * x5 * x6 * x7 +
        634965012726715944 * x6^2 * x7 +
        533272315529699460 * x2 * x7^2 +
        768264262397194680 * x3 * x7^2 +
        1439432472420278856 * x4 * x7^2 +
        1092285489784547448 * x5 * x7^2 +
        1762131427753049496 * x6 * x7^2 +
        1911645538035845040 * x7^3 +
        1598358037982940 * x2 * x4 +
        4962296840092020 * x2 * x5 +
        19574931887737200 * x3 * x5 - 4276082463884752 * x5^2 +
        6148498240368900 * x2 * x6 - 1575465283199340 * x3 * x6 -
        49598192718218744 * x4 * x6 - 73713482550000312 * x5 * x6 -
        46594541578129512 * x6^2 +
        1852749037172826 * x2 * x7 +
        46919833137162100 * x3 * x7 +
        8624665702928802 * x4 * x7 - 43918103184802248 * x5 * x7 -
        203505367093190000 * x6 * x7 - 441221259797163744 * x7^2 +
        376790373540003 * x2 +
        665358291872848 * x3 +
        2276643311479470 * x4 +
        2416863236704704 * x5 - 596248247189420 * x6 + 2108391119660688 * x7,
        402794071892552340000 * x3 * x6 * x7^3 +
        287710051351823100000 * x2 * x7^4 +
        387651437610877440000 * x3 * x7^4 +
        239253621650463420000 * x4 * x7^4 - 155969133101251470000 * x5 * x7^4 -
        324052373627842860000 * x6 * x7^4 - 221082460512453540000 * x7^5 -
        274651828266816000 * x6^4 - 531801334144080000 * x6^3 * x7 +
        6057053712669960000 * x2 * x4 * x7^2 - 15396909092921070000 * x2 * x5 * x7^2 -
        65597841106263180000 * x3 * x5 * x7^2 - 44904433088012355000 * x2 * x6 * x7^2 -
        83453364109428007500 * x3 * x6 * x7^2 - 16766879887916613000 * x4 * x6 * x7^2 -
        2713691912308357500 * x6^2 * x7^2 + 24657981359780047500 * x2 * x7^3 -
        27829247378712507000 * x3 * x7^3 +
        103012214583573067500 * x4 * x7^3 +
        171667782184638987000 * x5 * x7^3 +
        205631041710109275000 * x6 * x7^3 +
        259918224825486474000 * x7^4 +
        2560458795832080000 * x2 * x4 * x6 +
        1794442703129107200 * x2 * x6^2 - 3522240182638597680 * x3 * x6^2 -
        1527182512783293600 * x4 * x6^2 - 2353626491514161760 * x5 * x6^2 -
        4651688095247044800 * x6^3 - 1590409895706450000 * x2 * x4 * x7 +
        14889259875285835650 * x2 * x5 * x7 +
        38923175382016702140 * x3 * x5 * x7 +
        24563109065761197360 * x2 * x6 * x7 +
        26184110124714210630 * x3 * x6 * x7 +
        2032150207910432220 * x4 * x6 * x7 +
        4402305350185749552 * x5 * x6 * x7 - 5974462908888193512 * x6^2 * x7 -
        33992361042014301030 * x2 * x7^2 - 48455663721503596740 * x3 * x7^2 -
        89547816005149263138 * x4 * x7^2 - 89998994666715967404 * x5 * x7^2 -
        107048997353187498258 * x6 * x7^2 - 133777448229162811320 * x7^3 -
        142546842625240170 * x2 * x4 - 276486117175690110 * x2 * x5 -
        1087626072825021600 * x3 * x5 + 274407100518894046 * x5^2 -
        1571087718900080550 * x2 * x6 - 2918880280283244030 * x3 * x6 -
        1637265834795660988 * x4 * x6 - 683662709926073424 * x5 * x6 -
        1940722710250787724 * x6^2 +
        513843254657337852 * x2 * x7 +
        1661146715241469850 * x3 * x7 +
        5736458656778558604 * x4 * x7 +
        6897719665051652604 * x5 * x7 +
        8740892078025451475 * x6 * x7 +
        23079545990815161462 * x7^2 +
        77050619565547731 * x2 +
        111056283617063846 * x3 +
        89699231148074565 * x4 +
        136769191109438508 * x5 +
        379746048216973010 * x6 +
        273791090696914326 * x7,
        12083822156776570200000 * x2 * x6 * x7^3 - 3997655450362173600000 * x2 * x7^4 -
        8740328507382752280000 * x3 * x7^4 - 17262603081109386000000 * x4 * x7^4 -
        14005422447121115010000 * x5 * x7^4 - 8495017832019618900000 * x6 * x7^4 -
        10760355920558183940000 * x7^5 +
        5721913088892000000 * x6^4 +
        13210887572883000000 * x6^3 * x7 - 885844105477981650000 * x2 * x4 * x7^2 -
        20334394606820580000 * x2 * x5 * x7^2 +
        1046194907646840120000 * x3 * x5 * x7^2 -
        1035940736546351415000 * x2 * x6 * x7^2 +
        632333534234876077500 * x3 * x6 * x7^2 +
        100542837350784465000 * x4 * x6 * x7^2 +
        84933989570626267500 * x6^2 * x7^2 - 2161461353595032677500 * x2 * x7^3 +
        616850671163638275000 * x3 * x7^3 +
        1316206714711188982500 * x4 * x7^3 +
        809138111009214285000 * x5 * x7^3 +
        14648228349442155000 * x6 * x7^3 +
        2715542457268281030000 * x7^4 - 247438027447453440000 * x2 * x4 * x6 -
        340194655260604684800 * x2 * x6^2 - 132801806181350900880 * x3 * x6^2 -
        288989393041325901600 * x4 * x6^2 - 298029877868574034560 * x5 * x6^2 -
        217902352807355018400 * x6^3 + 195942263855120055000 * x2 * x4 * x7 -
        212176411491336440850 * x2 * x5 * x7 - 729772348670683669260 * x3 * x5 * x7 -
        490259710928489313240 * x2 * x6 * x7 - 750109749208544615670 * x3 * x6 * x7 -
        323891019243539062380 * x4 * x6 * x7 - 563708640362268206988 * x5 * x6 * x7 -
        641448118668807405072 * x6^2 * x7 +
        738279185983368741270 * x2 * x7^2 +
        722736118945520430660 * x3 * x7^2 +
        1329591580238859346422 * x4 * x7^2 +
        1072296779955925028076 * x5 * x7^2 +
        893335121644625385102 * x6 * x7^2 +
        1356028349081613531480 * x7^3 +
        9746478605721907530 * x2 * x4 +
        8693860865308920990 * x2 * x5 +
        25000696910908976400 * x3 * x5 - 6846158557915467074 * x5^2 +
        31019576471692280550 * x2 * x6 +
        57490246426702460670 * x3 * x6 +
        67933172457734798972 * x4 * x6 +
        42572415345747588456 * x5 * x6 +
        146941707088951370556 * x6^2 +
        28233862251443384562 * x2 * x7 +
        38030511624100863950 * x3 * x7 - 57543086770457509326 * x4 * x7 -
        46819063691422475676 * x5 * x7 + 66067534604104101875 * x6 * x7 -
        307894795061583914778 * x7^2 - 2366242092792301764 * x2 -
        4892556797495340874 * x3 - 6337280699840078985 * x4 - 7896764206321738452 * x5 -
        15956239459695917290 * x6 - 15769964522449570794 * x7,
        12083822156776570200000 * x3 * x5 * x7^3 - 15445486967308398000000 * x2 * x7^4 -
        14209848009923726160000 * x3 * x7^4 - 14718640521788002800000 * x4 * x7^4 -
        6023739917250275220000 * x5 * x7^4 - 6269050592613408600000 * x6 * x7^4 -
        7771199913355558680000 * x7^5 +
        4119777424002240000 * x6^4 +
        5418988278303600000 * x6^3 * x7 +
        763366916435022900000 * x2 * x4 * x7^2 -
        335966900999403960000 * x2 * x5 * x7^2 -
        2845351008457013160000 * x3 * x5 * x7^2 -
        331596069408786780000 * x2 * x6 * x7^2 -
        532602966330879795000 * x3 * x6 * x7^2 +
        737938280062210500000 * x4 * x6 * x7^2 +
        343030594051472535000 * x6^2 * x7^2 +
        4368060555313856895000 * x2 * x7^3 +
        3334536855581406930000 * x3 * x7^3 +
        3707172680791413015000 * x4 * x7^3 +
        1256082522493999590000 * x5 * x7^3 +
        1615944540815938710000 * x6 * x7^3 +
        3537996104899830600000 * x7^4 +
        219712879431590220000 * x2 * x4 * x6 +
        347320532757450002400 * x2 * x6^2 +
        283520772284154949440 * x3 * x6^2 +
        493523992419698260800 * x4 * x6^2 +
        522495336106318049280 * x5 * x6^2 +
        490317651201232015200 * x6^3 - 60776388304803990000 * x2 * x4 * x7 -
        81416676588074462700 * x2 * x5 * x7 - 49767148934887686120 * x3 * x5 * x7 +
        131791412844999456120 * x2 * x6 * x7 +
        415979861113841354460 * x3 * x6 * x7 +
        567442427604867865440 * x4 * x6 * x7 +
        792051367207602800844 * x5 * x6 * x7 +
        1257288148903649733936 * x6^2 * x7 +
        176943338928086002740 * x2 * x7^2 +
        795027532690595032920 * x3 * x7^2 +
        1083031574369092872264 * x4 * x7^2 +
        1591264877766196573512 * x5 * x7^2 +
        2312894217233382830424 * x6 * x7^2 +
        973557623653719071760 * x7^3 - 11514320695753132140 * x2 * x4 -
        7051596606367075620 * x2 * x5 - 3531605458192063200 * x3 * x5 +
        1088627109292662712 * x5^2 - 14618889718562106900 * x2 * x6 -
        5902782954232013460 * x3 * x6 - 36700708937875806136 * x4 * x6 -
        31639545268370093928 * x5 * x6 - 122799155575463572728 * x6^2 -
        76764942596362306806 * x2 * x7 - 140898380527305786100 * x3 * x7 -
        185518768290663755262 * x4 * x7 - 257677639112137586712 * x5 * x7 -
        505674662579150899000 * x6 * x7 - 485387612421063734136 * x7^2 +
        1571704789911039807 * x2 +
        3680596414370818112 * x3 +
        6447579112961027430 * x4 +
        7810506451229555376 * x5 +
        14129005276541464520 * x6 +
        18526648409152050072 * x7,
        4833528862710628080000 * x2 * x5 * x7^3 +
        4651817251330529280000 * x2 * x7^4 +
        3852286161258094560000 * x3 * x7^4 +
        8940211279900860960000 * x4 * x7^4 +
        9430832630627127720000 * x5 * x7^4 +
        8019539115575027040000 * x6 * x7^4 +
        8128566082403086320000 * x7^5 - 274651828266816000 * x6^4 -
        2748762170154000000 * x6^3 * x7 - 609624641316370680000 * x2 * x4 * x7^2 -
        311888706054920280000 * x2 * x5 * x7^2 +
        596823348552312960000 * x3 * x5 * x7^2 -
        473462052748853220000 * x2 * x6 * x7^2 -
        591340447728745065000 * x3 * x6 * x7^2 -
        1534298094878931018000 * x4 * x6 * x7^2 + 123145770198690720000 * x6^2 * x7^2 -
        2473861932562439835000 * x2 * x7^3 - 3564744418952547012000 * x3 * x7^3 -
        4272839121853890315000 * x4 * x7^3 - 3177508461054450588000 * x5 * x7^3 -
        2499986772195438240000 * x6 * x7^3 - 4504747913378496576000 * x7^4 -
        105847932987910620000 * x2 * x4 * x6 - 149410745247416047200 * x2 * x6^2 -
        122635104797890264320 * x3 * x6^2 - 237593601610127942400 * x4 * x6^2 -
        234005271794796565440 * x5 * x6^2 - 206520429937192956000 * x6^3 +
        19409535685653750000 * x2 * x4 * x7 - 62490587845461246900 * x2 * x5 * x7 -
        144029237807334428640 * x3 * x5 * x7 - 132144171043537584360 * x2 * x6 * x7 -
        255091840584053983380 * x3 * x6 * x7 - 215287851324786640920 * x4 * x6 * x7 -
        348346930903877626812 * x5 * x6 * x7 - 546383500411439174328 * x6^2 * x7 +
        211337843799125901780 * x2 * x7^2 +
        213909432369838536240 * x3 * x7^2 +
        242456444314913998728 * x4 * x7^2 +
        11177581031691126624 * x5 * x7^2 - 253913242144326038352 * x6 * x7^2 +
        505763236946437161120 * x7^3 +
        2043685928422213920 * x2 * x4 +
        12608073366432579360 * x2 * x5 +
        14208453717623895600 * x3 * x5 - 431492285615938376 * x5^2 +
        11219171876049477600 * x2 * x6 +
        1448241401501439480 * x3 * x6 - 10833131736446703472 * x4 * x6 -
        19501722521918684256 * x5 * x6 +
        27273548001635901744 * x6^2 +
        33447758301984547638 * x2 * x7 +
        69005659312604769200 * x3 * x7 +
        75071123149520110326 * x4 * x7 +
        92152307276341219776 * x5 * x7 +
        122872680717711946100 * x6 * x7 +
        33571568277342052728 * x7^2 - 499508128592148261 * x2 -
        715337337620526976 * x3 + 71451531942169110 * x4 - 878269013884460448 * x5 -
        1932987493564374610 * x6 - 896590349058845256 * x7,
        2369376893485602000000 * x2 * x4 * x7^3 +
        1011884267293099200000 * x2 * x7^4 +
        351309115334857680000 * x3 * x7^4 +
        306415658405656800000 * x4 * x7^4 - 37589362746275340000 * x5 * x7^4 -
        983974313911188600000 * x6 * x7^4 - 611762424979665960000 * x7^5 -
        1099504868061600000 * x6^4 - 2421528578469000000 * x6^3 * x7 -
        424814640868718100000 * x2 * x4 * x7^2 + 78212741522624280000 * x2 * x5 * x7^2 -
        74098522636229520000 * x3 * x5 * x7^2 - 334059719538440610000 * x2 * x6 * x7^2 -
        183668881650911865000 * x3 * x6 * x7^2 -
        328738014178045290000 * x4 * x6 * x7^2 + 23857071762348045000 * x6^2 * x7^2 -
        25725141459383535000 * x2 * x7^3 +
        75043106792813250000 * x3 * x7^3 +
        622026484278213105000 * x4 * x7^3 +
        953003999913225390000 * x5 * x7^3 +
        1138610935790687970000 * x6 * x7^3 +
        1288386048598138620000 * x7^4 +
        76931549111771640000 * x2 * x4 * x6 +
        93089177940065860800 * x2 * x6^2 +
        55145291020882152480 * x3 * x6^2 +
        118245393693851673600 * x4 * x6^2 +
        110592527024857793760 * x5 * x6^2 +
        87803584812492026400 * x6^3 +
        10566066119451570000 * x2 * x4 * x7 +
        35374072746746249100 * x2 * x5 * x7 +
        119631032478912459960 * x3 * x5 * x7 +
        180544767933061715040 * x2 * x6 * x7 +
        226401178904067428820 * x3 * x6 * x7 +
        185070099522333921480 * x4 * x6 * x7 +
        283163884873297394148 * x5 * x6 * x7 +
        334686633426358637112 * x6^2 * x7 - 55988240863925946420 * x2 * x7^2 -
        119792523772873584360 * x3 * x7^2 - 229749390551360456712 * x4 * x7^2 -
        250371228702763921896 * x5 * x7^2 - 64775059332995977992 * x6 * x7^2 -
        517850345694571840080 * x7^3 - 3571139321517823380 * x2 * x4 -
        1880719572936276540 * x2 * x5 - 5669545583150654400 * x3 * x5 +
        1858744544625527504 * x5^2 - 18008700907244100300 * x2 * x6 -
        23786050235761763820 * x3 * x6 - 30476673396691911512 * x4 * x6 -
        28763392942685096376 * x5 * x6 - 62975901069482267976 * x6^2 -
        7751494517057823402 * x2 * x7 - 3485488328949736700 * x3 * x7 +
        9976471079622035646 * x4 * x7 +
        15052594357806251496 * x5 * x7 - 64223566565870015000 * x6 * x7 +
        37345833958967443488 * x7^2 +
        254326616933412369 * x2 +
        1053437982939237304 * x3 +
        1616680834458024810 * x4 +
        1849348830711404592 * x5 +
        4792773805049428840 * x6 +
        4924936710423156624 * x7,
        2345368650380496000 * x6^3 * x7^2 - 4690737300760992000 * x4 * x7^4 -
        19750472845309440000 * x5 * x7^4 - 23206805593238592000 * x6 * x7^4 -
        21643226492984928000 * x7^5 - 21843058203532800 * x6^4 -
        27015322934808000 * x6^3 * x7 - 123440455283184000 * x2 * x6 * x7^2 -
        460916553791304000 * x3 * x6 * x7^2 - 1044963581674334400 * x4 * x6 * x7^2 -
        1008047403068472000 * x6^2 * x7^2 +
        290915225024112000 * x2 * x7^3 +
        1132713736762982400 * x3 * x7^3 +
        4221112048951824000 * x4 * x7^3 +
        7925013493944393600 * x5 * x7^3 +
        8229644752364568000 * x6 * x7^3 +
        13566723017025859200 * x7^4 - 122159954592748800 * x2 * x6^2 -
        284691166828323840 * x3 * x6^2 - 445628584042012800 * x4 * x6^2 -
        509982430568607360 * x5 * x6^2 - 571879943403556320 * x6^3 -
        75074896780416000 * x2 * x4 * x7 +
        102094688062598400 * x2 * x5 * x7 +
        122916071128144320 * x3 * x5 * x7 +
        19618992417660480 * x2 * x6 * x7 +
        114196856249867040 * x3 * x6 * x7 - 123302694438256320 * x4 * x6 * x7 -
        58536721258353468 * x5 * x6 * x7 - 594648140188664952 * x6^2 * x7 -
        319300087104311040 * x2 * x7^2 - 1007575333889267520 * x3 * x7^2 -
        1895806577544146748 * x4 * x7^2 - 2248419757742396784 * x5 * x7^2 -
        2728807526321084988 * x6 * x7^2 - 3407427756066054240 * x7^3 +
        2222589288479040 * x2 * x4 +
        3175364207826720 * x2 * x5 +
        6974581705509600 * x3 * x5 - 2032033327050004 * x5^2 -
        947294019247680 * x2 * x6 - 149462766704160 * x3 * x6 -
        2154139424738168 * x4 * x6 +
        17444217557560296 * x5 * x6 +
        45685833570629976 * x6^2 +
        31825202243403582 * x2 * x7 +
        107191020809056720 * x3 * x7 +
        152809130503251894 * x4 * x7 +
        210162324598759824 * x5 * x7 +
        402724642905910030 * x6 * x7 +
        450809877847740492 * x7^2 - 1174726723748829 * x2 - 2628874195291724 * x3 -
        3400338589448460 * x4 - 4260784854541032 * x5 - 6652845714562160 * x6 -
        6937771054162404 * x7,
        11179590566813697600000 * x6^4 * x7 +
        17887344906901916160000 * x3 * x7^4 +
        112972704675169996800000 * x4 * x7^4 +
        185816563210513984320000 * x5 * x7^4 +
        140223677759601321600000 * x6 * x7^4 +
        142544971487199790080000 * x7^5 +
        242788168031666400000 * x6^4 - 4781562161063945400000 * x6^3 * x7 +
        1412158808439624960000 * x2 * x5 * x7^2 +
        5272885375372517760000 * x3 * x5 * x7^2 +
        6931334006955895680000 * x2 * x6 * x7^2 +
        15884107613929685520000 * x3 * x6 * x7^2 +
        25498796123560799520000 * x4 * x6 * x7^2 -
        5093923179851283960000 * x6^2 * x7^2 - 3211528898683095120000 * x2 * x7^3 -
        3140710906246711200000 * x3 * x7^3 - 46119254282109159840000 * x4 * x7^3 -
        87846689734539905520000 * x5 * x7^3 - 92861617350306130560000 * x6 * x7^3 -
        126530563253077278360000 * x7^4 - 429428409583979520000 * x2 * x4 * x6 +
        272416749549622329600 * x2 * x6^2 +
        2317630231410244133760 * x3 * x6^2 +
        3828487316759610523200 * x4 * x6^2 +
        3898951704973396113120 * x5 * x6^2 +
        4299842159637491656800 * x6^3 +
        756600285343163040000 * x2 * x4 * x7 - 284208510201097060800 * x2 * x5 * x7 -
        406160687158388772480 * x3 * x5 * x7 +
        294066711374105292480 * x2 * x6 * x7 +
        1594330563723604611840 * x3 * x6 * x7 +
        4317233485049475971760 * x4 * x6 * x7 +
        4756973792268693905676 * x5 * x6 * x7 +
        7463186473195456219944 * x6^2 * x7 +
        2069103807017360436960 * x2 * x7^2 +
        5649456852277853359680 * x3 * x7^2 +
        10849555256551008457356 * x4 * x7^2 +
        16445481263218793090448 * x5 * x7^2 +
        25700536588769205717996 * x6 * x7^2 +
        35989327139158269995040 * x7^3 +
        26440058865854125440 * x2 * x4 - 45710502715545810480 * x2 * x5 -
        98569730615956282800 * x3 * x5 +
        148505349210041505748 * x5^2 +
        251756595498693776400 * x2 * x6 +
        253097614014278258160 * x3 * x6 +
        330503978293650922256 * x4 * x6 +
        257677724866938707688 * x5 * x6 - 286461245597223136512 * x6^2 -
        405380608948368591174 * x2 * x7 - 1642525211719619070400 * x3 * x7 -
        2068529386854638379198 * x4 * x7 - 2043586976184299845248 * x5 * x7 -
        3615557113664159173750 * x6 * x7 - 3344635899538797395244 * x7^2 -
        223414626088673247 * x2 + 4812363088524263348 * x3 - 4049774217624111780 * x4 +
        8103444364272553704 * x5 +
        25726026814572573080 * x6 +
        42567201322903021188 * x7,
        8940094984469577696768000 * x6^5 - 10217251410822374510592000 * x2 * x7^4 -
        86040011880609469562880000 * x3 * x7^4 -
        183910525394802741190656000 * x4 * x7^4 -
        58796644148192224834560000 * x5 * x7^4 +
        36588093287465056296960000 * x6 * x7^4 +
        12663487042722055507968000 * x7^5 - 4259193218827295856307200 * x6^4 -
        4085860158886145865825600 * x6^3 * x7 -
        1344375185634522961920000 * x2 * x4 * x7^2 -
        6473008951636865324544000 * x2 * x5 * x7^2 -
        23105889800127693950976000 * x3 * x5 * x7^2 -
        24120491736081188597760000 * x2 * x6 * x7^2 -
        39928464875513491636320000 * x3 * x6 * x7^2 -
        47049632359913490684249600 * x4 * x6 * x7^2 +
        11721233862012944441577600 * x6^2 * x7^2 +
        9555666158794624708320000 * x2 * x7^3 - 5043392641276411255142400 * x3 * x7^3 +
        42685913439261071464867200 * x4 * x7^3 +
        70544274457162836659097600 * x5 * x7^3 +
        84508022796154581139027200 * x6 * x7^3 +
        81196517384109610072012800 * x7^4 +
        1075210032609436449024000 * x2 * x4 * x6 +
        2100512140990435217260800 * x2 * x6^2 +
        2251262315281813956645120 * x3 * x6^2 +
        4219181369923992400281600 * x4 * x6^2 +
        6041545126356615693030720 * x5 * x6^2 +
        5407522202710514194420320 * x6^3 +
        741250895943093262272000 * x2 * x4 * x7 -
        797813721271896747350400 * x2 * x5 * x7 +
        988808369484063132426240 * x3 * x5 * x7 +
        2787741776140595914685760 * x2 * x6 * x7 +
        1128224323626088312896480 * x3 * x6 * x7 +
        3390029013367678657060800 * x4 * x6 * x7 +
        5408939572690505817685356 * x5 * x6 * x7 +
        14180884134673292753145624 * x6^2 * x7 +
        699298214779406616491520 * x2 * x7^2 +
        8313318763266993470991360 * x3 * x7^2 +
        11665487410194525425391276 * x4 * x7^2 +
        5849276255475256205295408 * x5 * x7^2 +
        10256378100084708010907436 * x6 * x7^2 - 20459502795533743025828640 * x7^3 -
        162815864031604072523520 * x2 * x4 +
        49424804205091519033440 * x2 * x5 +
        47988312855330582633600 * x3 * x5 - 150693843939162687985372 * x5^2 -
        878189273528044927475520 * x2 * x6 - 1472683255672218136475760 * x3 * x6 -
        2112337225431214244622104 * x4 * x6 - 2787780304366851274392552 * x5 * x6 -
        3020888960100617344928232 * x6^2 - 135929698554130086104694 * x2 * x7 +
        663571163647517821728880 * x3 * x7 +
        465338185474223476280082 * x4 * x7 - 654879033647188375104048 * x5 * x7 -
        5431193858999720703809030 * x6 * x7 - 4145516130762164326480764 * x7^2 +
        60188029648557928589793 * x2 +
        128988138387387953705548 * x3 +
        220293049552999951753020 * x4 +
        280421333998426596750984 * x5 +
        421677225079343644739800 * x6 +
        491499030806558497820148 * x7,
        2755111451745058005600000 * x6 * x7^5 + 2886307235161489339200000 * x7^6 -
        65597891708215666800000 * x2 * x7^4 - 244898795710671822720000 * x3 * x7^4 -
        520499931397480148400000 * x4 * x7^4 - 898490800983342764340000 * x5 * x7^4 -
        1403071785714974146200000 * x6 * x7^4 - 2086917764773479426360000 * x7^5 -
        15024443171502885120000 * x6^4 - 25477907298285138840000 * x6^3 * x7 -
        98396837562323500200000 * x2 * x4 * x7^2 -
        40786901761038224220000 * x2 * x5 * x7^2 -
        257048363604261242520000 * x3 * x5 * x7^2 -
        371086921010579264010000 * x2 * x6 * x7^2 -
        500188545095939785215000 * x3 * x6 * x7^2 -
        685747931701420019250000 * x4 * x6 * x7^2 +
        37879102290859368435000 * x6^2 * x7^2 +
        135973767493434919815000 * x2 * x7^3 +
        16160347008377656110000 * x3 * x7^3 +
        847562310968868630135000 * x4 * x7^3 +
        1674596847442064730210000 * x5 * x7^3 +
        1944679556375893870650000 * x6 * x7^3 +
        2064599795375093094300000 * x7^4 - 2073411323004630960000 * x2 * x4 * x6 -
        9301837679935410355200 * x2 * x6^2 - 18164451790885277049120 * x3 * x6^2 -
        27164912406072601118400 * x4 * x6^2 - 31543131958707812077440 * x5 * x6^2 -
        46603925393795102829600 * x6^3 +
        6047910399873345120000 * x2 * x4 * x7 +
        10065520825097297192100 * x2 * x5 * x7 +
        33311947764799562306760 * x3 * x5 * x7 +
        31692586716004239408240 * x2 * x6 * x7 +
        36843091658752671309420 * x3 * x6 * x7 +
        26527786781400014717880 * x4 * x6 * x7 +
        58018791072545905067388 * x5 * x6 * x7 +
        29000068732348953573672 * x6^2 * x7 - 7265241264303463435020 * x2 * x7^2 -
        26545919748898140803160 * x3 * x7^2 - 153639904886163377161272 * x4 * x7^2 -
        287821505210596105124376 * x5 * x7^2 - 305184456083333826236952 * x6 * x7^2 -
        684032048478109287618480 * x7^3 - 123050397489832488780 * x2 * x4 +
        702697628791512055260 * x2 * x5 +
        552114994493879013600 * x3 * x5 +
        215105003326848599024 * x5^2 - 7369625729035717161300 * x2 * x6 -
        14235761458980393492420 * x3 * x6 - 18698099426609759153072 * x4 * x6 -
        21004017196562435261256 * x5 * x6 - 19223582515838658653856 * x6^2 +
        4397483263000063873338 * x2 * x7 +
        22954349212160988490300 * x3 * x7 +
        30429486873360666799626 * x4 * x7 +
        35736603981630324977976 * x5 * x7 +
        17356491003948263137000 * x6 * x7 +
        34857847506946779200928 * x7^2 +
        161479374871054654839 * x2 +
        213715114189230974824 * x3 +
        461109660661812002610 * x4 +
        737756984153345343552 * x5 +
        1292341020558589342540 * x6 +
        1804228648680546261744 * x7,
        676694040779487931200000 * x5 * x7^5 - 193341154508425123200000 * x7^6 -
        5330207267149564800000 * x2 * x7^4 - 24873450467229524160000 * x3 * x7^4 -
        65539446453093412800000 * x4 * x7^4 - 185353175830616113320000 * x5 * x7^4 -
        71302148608147657200000 * x6 * x7^4 - 762551183195081280000 * x7^5 +
        2931944405146716960000 * x6^4 +
        9077183849240587080000 * x6^3 * x7 +
        30886192050106793400000 * x2 * x4 * x7^2 -
        19128062169707736960000 * x2 * x5 * x7^2 +
        16136620669231161840000 * x3 * x5 * x7^2 +
        54220827994315896420000 * x2 * x6 * x7^2 +
        41587675274431656630000 * x3 * x6 * x7^2 +
        99139967100752137560000 * x4 * x6 * x7^2 +
        324212149319232330000 * x6^2 * x7^2 - 25489087663137016230000 * x2 * x7^3 -
        7827273022423179780000 * x3 * x7^3 - 121310506279407078270000 * x4 * x7^3 -
        247833200783902895460000 * x5 * x7^3 - 258156420240599840700000 * x6 * x7^3 -
        222216402263248064880000 * x7^4 +
        948872382261635520000 * x2 * x4 * x6 +
        1938182757738775142400 * x2 * x6^2 +
        2545830534250386901440 * x3 * x6^2 +
        3873091553101983700800 * x4 * x6^2 +
        6278342070112457885280 * x5 * x6^2 +
        10229191206668538343200 * x6^3 - 1077788657771492940000 * x2 * x4 * x7 -
        672428434454086120200 * x2 * x5 * x7 - 4267614202363599207120 * x3 * x5 * x7 -
        2468309912830195742880 * x2 * x6 * x7 - 820541799212953910040 * x3 * x6 * x7 +
        3259378129463001739440 * x4 * x6 * x7 - 5801758175644170277356 * x5 * x6 * x7 +
        3545688379725244772136 * x6^2 * x7 +
        1979018971877426277240 * x2 * x7^2 +
        10983211324866499183920 * x3 * x7^2 +
        32951693440101567981564 * x4 * x7^2 +
        56502794700953220102912 * x5 * x7^2 +
        60709778261751540749724 * x6 * x7^2 +
        112153295177293024661760 * x7^3 - 353529492775977131640 * x2 * x4 -
        1784554912875894120 * x2 * x5 +
        153327143733089086800 * x3 * x5 +
        36294481469143232612 * x5^2 +
        650179319214450612600 * x2 * x6 +
        1621496720101473554040 * x3 * x6 +
        1737892202248319643664 * x4 * x6 +
        2225814184045583072472 * x5 * x6 +
        1668024548175517296672 * x6^2 - 1147814602067561362506 * x2 * x7 -
        4268242773875123456600 * x3 * x7 - 5824354707074960736162 * x4 * x7 -
        7427169815531405372112 * x5 * x7 - 7436283258477930569750 * x6 * x7 -
        10950170731217636515836 * x7^2 +
        18395664193398382707 * x2 +
        50800165448786761012 * x3 +
        111588488950705009680 * x4 +
        82821588443053821576 * x5 +
        186977358438172326520 * x6 +
        223873612629464373972 * x7,
        131143305103064761066560000 * x4 * x7^5 + 40408301292260850748800000 * x7^6 -
        2512199369652141929760000 * x2 * x7^4 - 11434748280926857286400000 * x3 * x7^4 -
        41277876151466600886240000 * x4 * x7^4 -
        13564814188900364109720000 * x5 * x7^4 - 9337097705405655865680000 * x6 * x7^4 -
        23493223686933807927120000 * x7^5 - 154651968689472334080000 * x6^4 -
        904693880732059329912000 * x6^3 * x7 -
        2506033288972643910480000 * x2 * x4 * x7^2 +
        5327761963217006930760000 * x2 * x5 * x7^2 +
        564217223278227077520000 * x3 * x5 * x7^2 -
        2356920106079648875740000 * x2 * x6 * x7^2 +
        6031716659649530098470000 * x3 * x6 * x7^2 +
        5311280852744566829220000 * x4 * x6 * x7^2 -
        2582280979808647040118000 * x6^2 * x7^2 +
        8902883434803239345490000 * x2 * x7^3 +
        16352603138729083639140000 * x3 * x7^3 +
        16039629192101815334034000 * x4 * x7^3 +
        14355523876145786176284000 * x5 * x7^3 +
        7585995408578334461844000 * x6 * x7^3 +
        10635515422849297735560000 * x7^4 - 97454698209681766200000 * x2 * x4 * x6 -
        79629797484180988574400 * x2 * x6^2 - 50794521358296546267840 * x3 * x6^2 +
        595985683018815958003200 * x4 * x6^2 +
        244461174411671064169920 * x5 * x6^2 - 131102632427892149887200 * x6^3 -
        149723908207957596240000 * x2 * x4 * x7 -
        69810728225535085393800 * x2 * x5 * x7 +
        238132818505899458734320 * x3 * x5 * x7 -
        169713628262491170892320 * x2 * x6 * x7 -
        183553429191202259593560 * x3 * x6 * x7 -
        1267799012466176713713840 * x4 * x6 * x7 -
        175159235347731413280684 * x5 * x6 * x7 - 705340308552777961699896 * x6^2 * x7 -
        699926002145812934261640 * x2 * x7^2 - 2623815638748422126583120 * x3 * x7^2 -
        3458226913368560214360804 * x4 * x7^2 - 3337755152785155496940832 * x5 * x7^2 -
        3741333828106321356330564 * x6 * x7^2 - 5156346182740012202679360 * x7^3 +
        87905353113080506238040 * x2 * x4 - 21015412513775591656680 * x2 * x5 -
        52566646604988875860800 * x3 * x5 - 11910758606743831473532 * x5^2 +
        142963189486971495899400 * x2 * x6 +
        154262017092437729593560 * x3 * x6 +
        270140558654485351741096 * x4 * x6 +
        236532122026388960476008 * x5 * x6 +
        208527847299173086682808 * x6^2 +
        74283947898623382534966 * x2 * x7 - 1108137280306685827400 * x3 * x7 +
        45013053968152169775582 * x4 * x7 +
        187817790529342334656032 * x5 * x7 +
        691730883317631934902250 * x6 * x7 +
        1193385145621286498158596 * x7^2 - 10083748747131810001077 * x2 -
        13181186628077305152932 * x3 - 36042321215831696355480 * x4 -
        35666875635415993199736 * x5 - 72714390106558639868720 * x6 -
        95025180652045199448492 * x7,
        11709223669916496523800000 * x3 * x7^5 - 2459920939058087505000000 * x7^6 -
        380655540572743636200000 * x2 * x7^4 - 2783509948076911113060000 * x3 * x7^4 +
        372379009321073898000000 * x4 * x7^4 +
        1254770377859486789730000 * x5 * x7^4 +
        1125826736807434275900000 * x6 * x7^4 +
        1862705393862785978220000 * x7^5 +
        22703750068891011120000 * x6^4 +
        74289206809519132860000 * x6^3 * x7 - 37387164041455328100000 * x2 * x4 * x7^2 +
        545647307944204091340000 * x2 * x5 * x7^2 +
        2636096182996303769040000 * x3 * x5 * x7^2 +
        2227915144431528482970000 * x2 * x6 * x7^2 +
        2792617402892438378205000 * x3 * x6 * x7^2 +
        1923322768939622184615000 * x4 * x6 * x7^2 -
        258153040638932837287500 * x6^2 * x7^2 - 1836959525885646438705000 * x2 * x7^3 -
        1303980329590842115485000 * x3 * x7^3 - 3607438637880810259605000 * x4 * x7^3 -
        5548493813266284913890000 * x5 * x7^3 - 6782330290193358896385000 * x6 * x7^3 -
        6736431155964769696320000 * x7^4 +
        3474544081001819670000 * x2 * x4 * x6 +
        11218377375746353004400 * x2 * x6^2 +
        109566160464167370980640 * x3 * x6^2 +
        59213192719962106144800 * x4 * x6^2 +
        52774585558754583541680 * x5 * x6^2 +
        96148836512214411463200 * x6^3 +
        14999470921437042397500 * x2 * x4 * x7 -
        65060897175903220733700 * x2 * x5 * x7 -
        284953876065174838586220 * x3 * x5 * x7 -
        197546400835771833768780 * x2 * x6 * x7 -
        280212184176617654877240 * x3 * x6 * x7 -
        95299461042113966400360 * x4 * x6 * x7 -
        161831573158074625799136 * x5 * x6 * x7 - 67273342999258983485184 * x6^2 * x7 +
        89274658056479037010440 * x2 * x7^2 +
        48062609139776259314520 * x3 * x7^2 +
        559297304876309460019884 * x4 * x7^2 +
        761960127291416669516472 * x5 * x7^2 +
        931598861967440273275344 * x6 * x7^2 +
        2423554858953117414214560 * x7^3 +
        3096829111010280191910 * x2 * x4 - 3434763247012176874470 * x2 * x5 -
        2718613367900473252950 * x3 * x5 - 1020458922153876501328 * x5^2 +
        19730333788087551062850 * x2 * x6 +
        40834694078304015236490 * x3 * x6 +
        55123511449159483078984 * x4 * x6 +
        60164881779312679711332 * x5 * x6 +
        55202136314762527873032 * x6^2 - 3245176408048760227536 * x2 * x7 -
        39942581339718649283600 * x3 * x7 - 65290531151752543280472 * x4 * x7 -
        60325409391745552280922 * x5 * x7 + 4098373374657188968000 * x6 * x7 -
        64358353956064115802516 * x7^2 - 1011685625985017629383 * x2 -
        2320646184431820277553 * x3 - 3769910292728590265670 * x4 -
        4778751198022899582294 * x5 - 8685971954071407757505 * x6 -
        11206869794816131549668 * x7,
        6393236123774407101994800000 * x2 * x7^5 + 1515311298459781903080000000 * x7^6 -
        1092356626137293234718000000 * x2 * x7^4 +
        749246471963087291313120000 * x3 * x7^4 +
        1277754257554237338876000000 * x4 * x7^4 +
        1327677124187649266871840000 * x5 * x7^4 +
        1339338715187169326830200000 * x6 * x7^4 +
        769863374477413898441760000 * x7^5 - 7062929402554522617120000 * x6^4 -
        47737071299430524531520000 * x6^3 * x7 +
        710668297336002243638400000 * x2 * x4 * x7^2 -
        362806305987446359475580000 * x2 * x5 * x7^2 -
        996106170733829155592280000 * x3 * x5 * x7^2 +
        208838625720376349865960000 * x2 * x6 * x7^2 -
        112829090601530019379747500 * x3 * x6 * x7^2 +
        852609421581546160673505000 * x4 * x6 * x7^2 -
        136769060781284565198825000 * x6^2 * x7^2 +
        848172606859056606509947500 * x2 * x7^3 +
        1360108019646223407492660000 * x3 * x7^3 +
        888824889304010434627897500 * x4 * x7^3 +
        12459156444693045923940000 * x5 * x7^3 -
        199604887659440747065530000 * x6 * x7^3 - 465095436446213412374340000 * x7^4 +
        9129170070484867889460000 * x2 * x4 * x6 +
        59734933761963901318207200 * x2 * x6^2 +
        13021813598483579191492320 * x3 * x6^2 +
        19199881327312801576562400 * x4 * x6^2 +
        32667731676928410030951840 * x5 * x6^2 +
        16484350711483753187169600 * x6^3 - 40398257181734269194195000 * x2 * x4 * x7 -
        10444891368230404908094350 * x2 * x5 * x7 +
        73000284498591561460079640 * x3 * x5 * x7 -
        4421996657708926547677140 * x2 * x6 * x7 +
        47809136447963840460758130 * x3 * x6 * x7 +
        5014849571933248549763820 * x4 * x6 * x7 +
        33433608254049362615442732 * x5 * x6 * x7 +
        55591861753985973055897608 * x6^2 * x7 -
        62992650109339083214936530 * x2 * x7^2 +
        67814959580235779517365760 * x3 * x7^2 +
        11697571732557072753217242 * x4 * x7^2 +
        236312981572539691979009136 * x5 * x7^2 +
        242770630678297411274391222 * x6 * x7^2 +
        215185443618084832750787280 * x7^3 - 4327220481695280507225420 * x2 * x4 +
        1763728452343009468066140 * x2 * x5 +
        2147165259467299081475400 * x3 * x5 +
        650152272481695824104486 * x5^2 +
        5346118566658815053460300 * x2 * x6 +
        12462512905844946994065120 * x3 * x6 +
        15244953956809355822047292 * x4 * x6 +
        23212063076849437155467316 * x5 * x6 +
        21194139688896427353854916 * x6^2 - 14332664821561816688507718 * x2 * x7 -
        52749685777766515811887300 * x3 * x7 - 65356955577998062138493886 * x4 * x7 -
        85541420127204870019406736 * x5 * x7 - 60753627690555295760201125 * x6 * x7 -
        84151548930587395236322458 * x7^2 +
        212312398870574592881271 * x2 +
        684766942020229187664686 * x3 +
        1774296765154899525816915 * x4 +
        1996889296060073185141428 * x5 +
        4734882875090237823046685 * x6 +
        5626444410753850382035566 * x7,
        490487285309857856031052497456000000 * x7^7 -
        230332184169245603682474765528000000 * x7^6 +
        67606722290137285348757961897600000 * x2 * x7^4 +
        203476859404587414143723890304640000 * x3 * x7^4 +
        364802241816546725541495175430400000 * x4 * x7^4 +
        478089566314416212006471204263680000 * x5 * x7^4 +
        509726806292962092675026549563200000 * x6 * x7^4 +
        454251205279980690451113340015920000 * x7^5 +
        5967866400538906910229002820000000 * x6^4 +
        7675532399025941695555635725400000 * x6^3 * x7 +
        1294877498764641250339365685200000 * x2 * x4 * x7^2 +
        2827174304473390387376081947440000 * x2 * x5 * x7^2 +
        6226065093418532389943728502040000 * x3 * x5 * x7^2 +
        9400628863180557225938726465220000 * x2 * x6 * x7^2 +
        16065546447129988058871556447980000 * x3 * x6 * x7^2 +
        35399102390676089008834881561480000 * x4 * x6 * x7^2 +
        26088017212503339469638165455760000 * x6^2 * x7^2 -
        11646456407284585051703257848180000 * x2 * x7^3 -
        46891369182916638451378420331400000 * x3 * x7^3 -
        139559753356111318520834303383260000 * x4 * x7^3 -
        205157580341181328997323797406680000 * x5 * x7^3 -
        205769431879092277109251195656240000 * x6 * x7^3 -
        304814995372801409259950627725440000 * x7^4 +
        200010927710729147488480503720000 * x2 * x4 * x6 +
        407301864504298086139205050958400 * x2 * x6^2 -
        59420742140110347180970026096960 * x3 * x6^2 -
        257908412804004044527476657787200 * x4 * x6^2 -
        97454741870010032667128646911520 * x5 * x6^2 +
        339207546082849521498264034807200 * x6^3 +
        62966417579836747295208795360000 * x2 * x4 * x7 -
        380574995316177531467111164513200 * x2 * x5 * x7 -
        163549285845366083896768596253920 * x3 * x5 * x7 -
        1522530517734889587881283471796080 * x2 * x6 * x7 -
        4548742744949586901344489239510640 * x3 * x6 * x7 -
        7469843418697774038545750671254960 * x4 * x6 * x7 -
        11939705109268620035467027569015996 * x5 * x6 * x7 -
        16683293480265219931346927958798024 * x6^2 * x7 +
        714323689609980552531635492385840 * x2 * x7^2 +
        4742387983467818237051009955162720 * x3 * x7^2 +
        12064461073610139126363638715764724 * x4 * x7^2 +
        21724728039074529786704060987934192 * x5 * x7^2 +
        10396729520195074857073399208899284 * x6 * x7^2 +
        57324702791175745885248561223592160 * x7^3 +
        8771161156153704199558756823760 * x2 * x4 -
        197795322986199641692148607055920 * x2 * x5 -
        308674760743308253904341410931200 * x3 * x5 -
        137016009200243205590444927417108 * x5^2 +
        544880599941877902161483841975600 * x2 * x6 +
        1487460381258460364959775899850640 * x3 * x6 +
        2338888181668000450235936077851824 * x4 * x6 +
        2499644293726457786808999213160152 * x5 * x6 +
        2898688553140598788307735881688352 * x6^2 -
        29103531506215859973307888942146 * x2 * x7 -
        529779780740722076501106070141600 * x3 * x7 +
        205263867960413498063379472359558 * x4 * x7 -
        177778086061360723802715262573392 * x5 * x7 +
        3290583895677503205513069003743750 * x6 * x7 +
        606277615092370362716298353169324 * x7^2 -
        15772641642434959283774654187213 * x2 - 60197373549482648100492329161708 * x3 -
        137915100965321904395651572093620 * x4 -
        255031510221343654583843624334984 * x5 -
        415209119922756093620278917053680 * x6 - 615052252138533091728635488833348 * x7
    ]

    @test gb == map(f -> divexact(f, leading_coefficient(f)), sing_ans)
end
